# Copyright 2017-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas
from scipy.stats import norm


def _alphas(alpha: np.array, phi: float, t: np.array):
    """Alpha spending function."""
    pe = np.zeros(len(t))
    pd = np.zeros(len(t))
    for j, tt in enumerate(t):
        pe[j] = alpha * np.power(tt, phi)
        pd[j] = pe[j] if j == 0 else pe[j] - pe[j - 1]
    return pe, pd


def _qp(xq: float, last: float, nints: int, yam1: float, ybm1: float, stdv: float):
    hlast = (ybm1 - yam1) / nints
    grid = np.linspace(yam1, ybm1, nints + 1)
    fun = last * norm.cdf(grid, xq, stdv)
    qp = 0.5 * hlast * (2 * np.sum(fun) - fun[0] - fun[len(fun) - 1])
    return qp


def _bsearch(
    last: np.array,
    nints: int,
    pd: float,
    stdv: float,
    ya: float,
    yb: float,
) -> np.array:
    """
    Note: function signature slightly modified in comparison to R implementation (which takes complete nints
    array instead of scalar), but should be semantically equivalent
    """
    max_iter = 50
    tol = 1e-7
    de = 10
    uppr = yb
    q = _qp(uppr, last, nints, ya, yb, stdv)
    while abs(q - pd) > tol:
        de = de / 10
        temp = 1 if q > (pd + tol) else 0
        incr = 2 * temp - 1
        j = 1
        while j <= max_iter:
            uppr = uppr + incr * de
            q = _qp(uppr, last, nints, ya, yb, stdv)
            if abs(q - pd) > tol and j >= max_iter:
                break
            elif (incr == 1 and q <= (pd + tol)) or (incr == -1 and q >= (pd - tol)):
                j = max_iter

            j += 1
    ybval = uppr
    return ybval


_NORM_CONSTANT = 1 / np.sqrt(2 * np.pi)


def _fast_norm_pdf_prescaled(x: np.array, scale):
    norm_constant2 = _NORM_CONSTANT / scale
    pdf_val = norm_constant2 * np.exp(-0.5 * np.power(x, 2))
    return pdf_val


def _fcab(last: np.array, nints: int, yam1: float, h: float, x: np.array, stdv: float):
    X, Y = np.meshgrid(x / stdv, (h * np.linspace(0, nints, nints + 1) + yam1) / stdv)
    scaled_x = Y - X
    pdf_prescaled = _fast_norm_pdf_prescaled(scaled_x, stdv)
    last_transposed = np.transpose(np.tile(last, len(x)).reshape(len(x), nints + 1))

    f = last_transposed * pdf_prescaled
    area = 0.5 * h * (2 * f.sum(0) - np.transpose(f[0, :]) - np.transpose(f[nints, :]))
    return area


# TODO use dataclass as soon as stets was migrated to Python 3.7
class ComputationState:
    """
    Internal state that can be fed into bounds(...). Whenever the internal state changes, a new ComputationState object
    will be created.

    It is not intended that other packages code operates on the attributes of this class because the internal
    structure may be changed anytime.
    """

    def __init__(self, df: pandas.DataFrame, last_fcab: np.array):
        if df is None or any(df["zb"].isnull()) or len(df) > 0 and last_fcab is None:
            raise ValueError()

        self._df = df
        self._last_fcab = last_fcab

    @property
    def df(self):
        # copy to avoid side effects
        return self._df.copy()

    @property
    def last_fcab(self):
        """fcab calculation referring to the last row of df"""

        # copy to avoid side effects
        return None if self._last_fcab is None else np.copy(self._last_fcab)

    def __eq__(self, other):
        if isinstance(other, ComputationState):
            return self._df.equals(other._df) and np.array_equal(self._last_fcab, other._last_fcab)
        return False


def landem(
    t: np.array,
    alpha: float,
    phi: float,
    ztrun: float,
    state: ComputationState,
    max_nints: int = None,
):
    """
    This function is a Python implementation of landem.R of ldbounds package.
    https://cran.r-project.org/web/packages/ldbounds/index.html
    Source code of that landem.R: https://github.com/cran/ldbounds/blob/master/R/landem.R

    After making any changes, please run test_compare_with_ldbounds.py to gain confidence that functionality is
    not broken.

    :param t: Monotonically increasing information ratios
    :param alpha: corrected alpha (other than R implementation, we do not modify alpha based on number of sides)
    :param phi: exponent used by alpha-sepending function
    :param ztrun: max value for truncating bounds
    :param state: state to build the computation upon
    :param max_nints: max value that internal nints parameter can take. Limiting this value reduces accuracy of the
        calculation but can lead to crucial performance improvement
    :return: A dataframe where the "zb" column contains the bounds and the i-th row reflects the results
    for the i-th information ratio from t
    """

    df = state.df  # reading the property will copy the df to avoid side effects
    last_fcab = state.last_fcab

    if len(t) <= len(df):
        # Simply return complete state and the existing result
        return df.iloc[: len(t)], state
    elif len(t) > len(df):
        # We reindex because appending rows *individually* to a DataFrame is expensive
        df = df.reindex(range(len(t)))

    h = 0.05
    zninf = -ztrun
    tol = 1e-7

    # t2 = t  # ldbounds:::bounds() rescales t2=t/t.max() by default. We omit this because impact on bounds unclear

    if df.isnull().all().all():
        # start at index 0 if df was not yet initialized
        start = 0
    else:
        # start at the first index where "zb" column is null (or at the very end if all "zb" values are not null)
        zb_null_arr = np.where(df["zb"].isnull())
        start = zb_null_arr[0][0] - 1 if len(zb_null_arr[0]) > 0 else len(df) - 1

    rangestart = start + 1
    if start == 0:
        df.loc[0, "stdv"] = np.sqrt(t[0])

    df.loc[start + 1 : len(t), "stdv"] = np.sqrt(t[start + 1 : len(t)] - t[start : len(t) - 1])

    df["pe"], df["pd"] = _alphas(alpha, phi, t)
    df.loc[start:, "sdproc"] = np.sqrt(t[start:])
    df.loc[start:, "information_ratio"] = t[start:]

    if df.isnull().all(axis=0)[0]:
        # this needs to be done only to compute the very first row
        if df.at[start, "pd"] < 1:
            df.at[start, "zb"] = norm.ppf(1 - df.at[start, "pd"])
            if df.at[start, "zb"] > ztrun:
                df.at[start, "zb"] = ztrun
                df.at[start, "pd"] = 1 - norm.cdf(df.at[start, "zb"])
                df.at[start, "pe"] = df.at[start, "pd"]
                if len(t) > 1:
                    df.at[1, "pd"] = df.at[start + 1, "pe"] - df.at[start, "pe"]
            df.at[start, "yb"] = df.at[start, "zb"] * df.at[start, "stdv"]

        df.at[start, "za"] = zninf
        df.at[start, "ya"] = df.at[start, "za"] * df.at[start, "stdv"]
        df.at[start, "nints"] = np.ceil((df.at[start, "yb"] - df.at[start, "ya"]) / (h * df.at[start, "stdv"]))

        grid = np.linspace(
            df.at[start, "ya"],
            df.at[start, "yb"],
            int(df.at[start, "nints"] + 1),
        )
        scaled_x = grid / df.at[start, "stdv"]
        last_fcab = _fast_norm_pdf_prescaled(scaled_x, df.at[start, "stdv"])

    if len(t) >= 2:
        for i in range(rangestart, len(t)):
            if df["information_ratio"][i] - df["information_ratio"][i - 1] <= 1e-5:
                # If information ratio difference between time steps is 0, re-use result calculated for the previous
                # time step. Normally, it means that no data was added. We have to catch this case because nints
                # becomes float("inf") and makes the procedure crash. We check against 10e-6 instead of against 0
                # because an almost-zero information gain can cause pretty big numerical inaccuracy in practice.
                df.iloc[i] = df.iloc[i - 1]
                continue

            # Possible error in spending function.  May be due to truncation.
            if df.at[i, "pd"] != 1.0:
                df.at[i, "pd"] = df.at[i, "pe"] - df.at[i - 1, "pe"]
            df.at[i, "pd"] = df.at[i, "pd"].clip(0, 1)

            if df.at[i, "pd"] < tol:
                df.at[i, "zb"] = -zninf
                if df.at[i, "zb"] > ztrun:
                    df.at[i, "zb"] = ztrun
                    df.at[i, "pd"] = _qp(
                        df.at[i, "zb"] * df.at[i, "sdproc"],
                        last_fcab,
                        df.at[i - 1, "nints"],
                        df.at[i - 1, "ya"],
                        df.at[i - 1, "yb"],
                        df.at[i, "stdv"],
                    )
                    df.at[i, "pe"] = df.at[i, "pd"] + df.at[i - 1, "pe"]

                df.at[i, "yb"] = df.at[i, "zb"] * df.at[i, "sdproc"]
            elif df.at[i, "pd"] == 1.0:
                df.at[i, "zb"] = 0.0
                df.at[i, "yb"] = 0.0
            elif tol <= df.at[i, "pd"] < 1:
                df.at[i, "yb"] = _bsearch(
                    last_fcab,
                    int(df.loc[i - 1]["nints"]),  # differs from R because we modified signature of bsearch
                    df.at[i, "pd"],
                    df.at[i, "stdv"],
                    df.at[i - 1, "ya"],
                    df.at[i - 1, "yb"],
                )

                df.at[i, "zb"] = df.at[i, "yb"] / df.at[i, "sdproc"]

                if df.at[i, "zb"] > ztrun:
                    df.at[i, "zb"] = ztrun
                    df.at[i, "pd"] = _qp(
                        df.at[i, "zb"] * df.at[i, "sdproc"],
                        last_fcab,
                        int(df.at[i - 1, "nints"]),
                        df.at[i - 1, "ya"],
                        df.at[i - 1, "yb"],
                        df.at[i, "stdv"],
                    )
                    df.at[i, "pe"] = df.at[i, "pd"] + df.at[i - 1, "pe"]

                df.at[i, "yb"] = df.at[i, "zb"] * df.at[i, "sdproc"]

            # in landem.R, the following two statements are in side==1 if clause
            df.at[i, "ya"] = zninf * df.at[i, "sdproc"]
            df.at[i, "za"] = zninf

            nints_calc = np.ceil((df.at[i, "yb"] - df.at[i, "ya"]) / (h * df.at[i, "stdv"]))
            df.at[i, "nints"] = nints_calc if max_nints is None or nints_calc < max_nints else max_nints

            if i < len(t):
                # in R implementation, i < len(t)-1. However we run until len(t) because that calculation will be
                # required if landem() is called again with df used as a starting point
                hlast = (df.at[i - 1, "yb"] - df.at[i - 1, "ya"]) / df.at[i - 1, "nints"]
                x = np.linspace(
                    df.at[i, "ya"],
                    df.at[i, "yb"],
                    int(df.at[i, "nints"] + 1),
                )
                last_fcab = _fcab(
                    last_fcab, int(df.at[i - 1, "nints"]), df.at[i - 1, "ya"], hlast, x, df.at[i, "stdv"]
                )
    return df, ComputationState(df, last_fcab)


# Simple type to return results in a structured way
class CalculationResult:
    def __init__(self, df: pandas.DataFrame, state: ComputationState):
        self._df = df
        self._state = state

    @property
    def df(self):
        return self._df

    @property
    def bounds(self):
        return self._df["zb"].values

    @property
    def state(self):
        return self._state


columns = ["za", "zb", "ya", "yb", "pe", "pd", "stdv", "sdproc", "nints", "information_ratio"]

# Initial state to be fed into bounds() to calculate sequential bounds from scratch
EMPTY_STATE = ComputationState(df=pandas.DataFrame(index=None, columns=columns, dtype=float), last_fcab=None)


def bounds(
    t: np.array,
    alpha: float,
    rho: float,
    ztrun: float,
    sides: int,
    state: ComputationState = EMPTY_STATE,
    max_nints=None,
) -> CalculationResult:
    """
    See landem() for parameter explanation

    :return: If a state is provided, returns a tuple of result and state. Otherwise, return only boundary result.
    """

    def get_input_str():
        return (
            f"input params: t={t}, alpha={alpha}, sides={sides}, rho={rho}, ztrun={ztrun},"
            f"state_df={state.df.to_json()}, state_fcab={state.last_fcab}, max_nints={max_nints}"
        )

    if any(t == 0.0):
        raise ValueError(f"Information ratio must must not be zero, {get_input_str()}")
    if any(t[i] > t[i + 1] for i in range(len(t) - 1)):
        raise ValueError(f"Information ratio must be monotonically increasing, {get_input_str()}")
    if not (sides == 1 or sides == 2):
        raise ValueError(f"sides must either be one a zero, {get_input_str()}")

    if state is None:
        state = EMPTY_STATE

    alph = alpha / sides

    df_result, new_state = landem(t, alph, rho, ztrun, state, max_nints)

    # guardrail check
    fixed_horizon_bound = norm.ppf(1 - alph)
    last_sequential_bound = df_result["zb"].values[-1]
    if fixed_horizon_bound > last_sequential_bound:
        raise Exception(
            f"Last bound ({last_sequential_bound}) is less conservative than fixed horizon bound "
            f"({fixed_horizon_bound}), {get_input_str()} "
        )

    return CalculationResult(df_result, new_state)



# Copyright 2017-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, Dict, Union

NUMERATOR = "numerator"
NUMERATOR_SUM_OF_SQUARES = "numerator_sum_of_squares"
DENOMINATOR = "denominator"
BOOTSTRAPS = "bootstraps"
INTERVAL_SIZE = "interval_size"
ALPHA = "alpha"
FINAL_EXPECTED_SAMPLE_SIZE = "final_expected_sample_size"
ORDINAL_GROUP_COLUMN = "ordinal_group_column"
MDE = "mde"
METHOD = "method_column"
CORRECTION_METHOD = "correction_method"
ABSOLUTE = "absolute"
NUMBER_OF_COMPARISONS = "number_of_comparisons"
TREATMENT_WEIGHTS = "treatment_weights"
IS_BINARY = "is_binary"
FEATURE = "feature"
FEATURE_SUMSQ = "feature_sumsq"
FEATURE_CROSS = "feature_cross"

POINT_ESTIMATE = "point_estimate"
VARIANCE = "variance"
CI_LOWER, CI_UPPER = "ci_lower", "ci_upper"
ADJUSTED_LOWER, ADJUSTED_UPPER = "adjusted ci_lower", "adjusted ci_upper"
CI_WIDTH = "ci_width"
DIFFERENCE = "difference"
P_VALUE = "p-value"
ADJUSTED_P = "adjusted p-value"
SFX1, SFX2 = "_1", "_2"
STD_ERR = "std_err"
Z_CRIT = "z_crit"
ALPHA = "alpha"
ADJUSTED_ALPHA = "adjusted_alpha"
ADJUSTED_ALPHA_POWER_SAMPLE_SIZE = "adjusted_alpha_power_sample_size"
POWER = "power"
POWERED_EFFECT = "powered_effect"
ADJUSTED_POWER = "adjusted_power"
IS_SIGNIFICANT = "is_significant"
REQUIRED_SAMPLE_SIZE = "required_sample_size"
REQUIRED_SAMPLE_SIZE_METRIC = "required_sample_size_for_metric"
OPTIMAL_KAPPA = "optimal_kappa"
OPTIMAL_WEIGHTS = "optimal_weigghts"
ORIGINAL_POINT_ESTIMATE = "original_point_estimate"
ORIGINAL_VARIANCE = "original_variance"
VARIANCE_REDUCTION = "variance_reduction_rate"

BONFERRONI = "bonferroni"
HOLM = "holm"
HOMMEL = "hommel"
SIMES_HOCHBERG = "simes-hochberg"
SIDAK = "sidak"
HOLM_SIDAK = "holm-sidak"
FDR_BH = "fdr_bh"
FDR_BY = "fdr_by"
FDR_TSBH = "fdr_tsbh"
FDR_TSBKY = "fdr_tsbky"

BONFERRONI_ONLY_COUNT_TWOSIDED = "bonferroni-only-count-twosided"
BONFERRONI_DO_NOT_COUNT_NON_INFERIORITY = "bonferroni-do-not-count-non-inferiority"
SPOT_1 = "spot-1-bonferroni"

SPOT_1_HOLM = "spot-1-holm"
SPOT_1_HOMMEL = "spot-1-hommel"
SPOT_1_SIMES_HOCHBERG = "spot-1-simes-hochberg"
SPOT_1_SIDAK = "spot-1-sidak"
SPOT_1_HOLM_SIDAK = "spot-1-holm-sidak"
SPOT_1_FDR_BH = "spot-1-fdr_bh"
SPOT_1_FDR_BY = "spot-1-fdr_by"
SPOT_1_FDR_TSBH = "spot-1-fdr_tsbh"
SPOT_1_FDR_TSBKY = "spot-1-fdr_tsbky"

CORRECTION_METHODS = [
    BONFERRONI,
    HOLM,
    HOMMEL,
    SIMES_HOCHBERG,
    SIDAK,
    HOLM_SIDAK,
    FDR_BH,
    FDR_BY,
    FDR_TSBH,
    FDR_TSBKY,
    BONFERRONI_ONLY_COUNT_TWOSIDED,
    BONFERRONI_DO_NOT_COUNT_NON_INFERIORITY,
    SPOT_1,
    SPOT_1_HOLM,
    SPOT_1_HOMMEL,
    SPOT_1_SIMES_HOCHBERG,
    SPOT_1_SIDAK,
    SPOT_1_HOLM_SIDAK,
    SPOT_1_FDR_BH,
    SPOT_1_FDR_BY,
    SPOT_1_FDR_TSBH,
    SPOT_1_FDR_TSBKY,
]

CORRECTION_METHODS_THAT_SUPPORT_CI = [
    BONFERRONI,
    HOLM,
    HOMMEL,
    SIMES_HOCHBERG,
    BONFERRONI_ONLY_COUNT_TWOSIDED,
    BONFERRONI_DO_NOT_COUNT_NON_INFERIORITY,
    SPOT_1,
    SPOT_1_HOLM,
    SPOT_1_HOMMEL,
    SPOT_1_SIMES_HOCHBERG,
    SPOT_1_SIDAK,
    SPOT_1_HOLM_SIDAK,
    SPOT_1_FDR_BH,
    SPOT_1_FDR_BY,
    SPOT_1_FDR_TSBH,
    SPOT_1_FDR_TSBKY,
]

CORRECTION_METHODS_THAT_REQUIRE_METRIC_INFO = [
    BONFERRONI_DO_NOT_COUNT_NON_INFERIORITY,
    SPOT_1,
    SPOT_1_HOLM,
    SPOT_1_HOMMEL,
    SPOT_1_SIMES_HOCHBERG,
    SPOT_1_SIDAK,
    SPOT_1_HOLM_SIDAK,
    SPOT_1_FDR_BH,
    SPOT_1_FDR_BY,
    SPOT_1_FDR_TSBH,
    SPOT_1_FDR_TSBKY,
]

CORRECTION_METHODS_THAT_DONT_REQUIRE_METRIC_INFO = list(
    set(CORRECTION_METHODS) - set(CORRECTION_METHODS_THAT_REQUIRE_METRIC_INFO)
)

NULL_HYPOTHESIS = "null_hypothesis"
ALTERNATIVE_HYPOTHESIS = "alternative_hypothesis"
NIM = "non-inferiority margin"
NIM_COLUMN_DEFAULT = "non_inferiority_margin"
PREFERRED_DIRECTION_COLUMN_DEFAULT = "preferred_direction"
INCREASE_PREFFERED = "increase"
DECREASE_PREFFERED = "decrease"
TWO_SIDED = "two-sided"
PREFERENCE = "preference"
PREFERENCE_TEST = "preference_used_in_test"
PREFERENCE_DICT = {"smaller": DECREASE_PREFFERED, "larger": INCREASE_PREFFERED, TWO_SIDED: TWO_SIDED}
NIM_TYPE = Union[Tuple[float, str], Dict[str, Tuple[float, str]], bool]
METHOD_COLUMN_NAME = "method_column_name"
CHI2 = "chi-squared"
TTEST = "t-test"
ZTEST = "z-test"
ZTESTLINREG = "z-test-linreg"
BOOTSTRAP = "bootstrap"
METHODS = [CHI2, TTEST, ZTEST, BOOTSTRAP, ZTESTLINREG]
REGRESSION_PARAM = "regression_parameters"



from typing import Union, Iterable, Tuple

from pandas import DataFrame

from spotify_confidence.analysis.frequentist.confidence_computers.sample_size_computer import SampleSizeComputer
from ..confidence_utils import (
    listify,
)
from ..constants import BONFERRONI, ZTEST, METHOD_COLUMN_NAME


class SampleSizeCalculator:
    def __init__(
        self,
        data_frame: DataFrame,
        point_estimate_column: str,
        var_column: str,
        is_binary_column: str,
        categorical_group_columns: Union[None, str, Iterable] = None,
        interval_size: float = 0.95,
        correction_method: str = BONFERRONI,
        metric_column=None,
        power: float = 0.8,
    ):
        self._sample_size_computer = SampleSizeComputer(
            data_frame=data_frame.assign(**{METHOD_COLUMN_NAME: ZTEST}),
            categorical_group_columns=listify(categorical_group_columns),
            interval_size=interval_size,
            correction_method=correction_method.lower(),
            metric_column=metric_column,
            power=power,
            point_estimate_column=point_estimate_column,
            var_column=var_column,
            is_binary_column=is_binary_column,
        )

    def sample_size(
        self,
        treatment_weights: Iterable,
        mde_column: str,
        nim_column: str,
        preferred_direction_column: str,
        final_expected_sample_size_column: str = None,
    ) -> DataFrame:
        """Args:
            treatment_weights (Iterable): The first weight is treated as control, the rest as treatment groups.
            mde_column (str): Name of column in source dataframe containing the minimum detectable effect sizes
            nim_column (str): Name of column in source dataframe containing the non-inferiority margins.
            preferred_direction_column (str): Name of column in source dataframe containing the preferred direction
            of the metric, which can be "increase", "decrease" or None, the latter meaning a two-sided test will be
            performed.
            final_expected_sample_size_column (str): Name of column in source dataframe containing an expected
            sample size. If this is given, a confidence interval width around the avg will be returned in the
            "ci_width" column
        Returns:
            Dataframe containing a column "required_sample_size_metric" containing the sample size ot the control vs
            treatment comparison that requires the largest sample size.

        One of mde or nim has to be set for each row of the dataframe. If mde is set, the null hypothesis is no
            difference between control and treatment, and the sample size returned will be enough to detect a relative
            difference of mde or and absolute difference of mde*avg. If nim is set the null hypothesis is -nim*avg
            if preferred_direction is "increase" and +nim*avg if preferred direction is "decrease"
        """
        return self._sample_size_computer.compute_sample_size(
            treatment_weights, mde_column, nim_column, preferred_direction_column, final_expected_sample_size_column
        )

    def optimal_weights_and_sample_size(
        self, sample_size_df: DataFrame, number_of_groups: int
    ) -> Tuple[Iterable, int]:
        """Args:
            sample_size_df (DataFrame): A data frame returned by the sample_size method of this class
            number_of_groups (int): Number of groups in the experiment, including control
        Returns:
            Tuple (list, int), where the list contains optimal weights and the int is the sample size that would be
            required if those weights were used
        """
        return self._sample_size_computer.compute_optimal_weights_and_sample_size(sample_size_df, number_of_groups)

    def powered_effect(
        self,
        treatment_weights: Iterable,
        mde_column: str,
        nim_column: str,
        preferred_direction_column: str,
        sample_size: int,
    ) -> DataFrame:
        """Args:
            treatment_weights (Iterable): The first weight is treated as control, the rest as treatment groups.
            mde_column (str): Name of column in source dataframe containing the minimum detectable effect sizes
            nim_column (str): Name of column in source dataframe containing the non-inferiority margins.
            preferred_direction_column (str): Name of column in source dataframe containing the preferred direction
            of the metric, which can be "increase", "decrease" or None, the latter meaning a two-sided test will be
            performed.
            sample_size (int): Total sample size across all groups to base the powered effect calculation on
        Returns:
            Dataframe containing a column "powered_effect" containing the powered effect of the control vs
            treatment comparison that requires the largest powered effect.

        Mde or nims are only needed for some multiple correction methods and are there to give results
        consistent with the sample size calculations, i.e. if you first calculate a sample size for a
        specific MDE and then calculate the powered effect for that sample size, the original MDE will be returned.
        """
        return self._sample_size_computer.compute_powered_effect(
            treatment_weights, mde_column, nim_column, preferred_direction_column, sample_size
        )



from typing import Tuple, Union, Dict

import numpy as np
from pandas import DataFrame, Series
from scipy import optimize
from scipy import stats as st
from scipy.stats.stats import _unequal_var_ttest_denom

from statsmodels.stats.weightstats import _zconfint_generic, _zstat_generic

from spotify_confidence.analysis.confidence_utils import power_calculation
from spotify_confidence.analysis.constants import (
    NUMERATOR,
    NUMERATOR_SUM_OF_SQUARES,
    DENOMINATOR,
    INTERVAL_SIZE,
    FINAL_EXPECTED_SAMPLE_SIZE,
    ORDINAL_GROUP_COLUMN,
    POINT_ESTIMATE,
    CI_LOWER,
    CI_UPPER,
    ADJUSTED_LOWER,
    ADJUSTED_UPPER,
    VARIANCE,
    NUMBER_OF_COMPARISONS,
    TWO_SIDED,
    SFX2,
    SFX1,
    STD_ERR,
    PREFERENCE_TEST,
    NULL_HYPOTHESIS,
    DIFFERENCE,
    ALPHA,
    IS_SIGNIFICANT,
    HOLM,
    SPOT_1_HOLM,
    HOMMEL,
    SIMES_HOCHBERG,
    SPOT_1_HOMMEL,
    SPOT_1_SIMES_HOCHBERG,
    NIM,
    ADJUSTED_ALPHA,
    ORIGINAL_POINT_ESTIMATE,
)
from spotify_confidence.analysis.frequentist.sequential_bound_solver import bounds


def sequential_bounds(t: np.array, alpha: float, sides: int, state: DataFrame = None):
    return bounds(t, alpha, rho=2, ztrun=8, sides=sides, max_nints=1000, state=state)


def point_estimate(df: DataFrame, **kwargs: Dict[str, str]) -> float:
    numerator = kwargs[NUMERATOR]
    denominator = kwargs[DENOMINATOR]
    if (df[denominator] == 0).any():
        raise ValueError("""Can't compute point estimate: denominator is 0""")
    return df[numerator] / df[denominator]


def variance(df: DataFrame, **kwargs: Dict[str, str]) -> float:
    numerator = kwargs[NUMERATOR]
    denominator = kwargs[DENOMINATOR]
    numerator_sumsq = kwargs[NUMERATOR_SUM_OF_SQUARES]
    binary = df[numerator_sumsq] == df[numerator]
    if binary.all():
        # This equals row[POINT_ESTIMATE]*(1-row[POINT_ESTIMATE]) when the data is binary,
        # and also gives a robust fallback in case it's not
        variance = df[numerator_sumsq] / df[denominator] - df[ORIGINAL_POINT_ESTIMATE] ** 2
    else:
        variance = (df[numerator_sumsq] - np.power(df[numerator], 2) / df[denominator]) / (df[denominator] - 1)
    if (variance < 0).any():
        raise ValueError("Computed variance is negative. " "Please check your inputs.")
    return variance


def std_err(df: Series, **kwargs: Dict[str, str]) -> float:
    denominator = kwargs[DENOMINATOR]
    return np.sqrt(df[VARIANCE + SFX1] / df[denominator + SFX1] + df[VARIANCE + SFX2] / df[denominator + SFX2])


def add_point_estimate_ci(df: Series, **kwargs: Dict[str, str]) -> Series:
    denominator = kwargs[DENOMINATOR]
    interval_size = kwargs[INTERVAL_SIZE]
    df[CI_LOWER], df[CI_UPPER] = _zconfint_generic(
        mean=df[POINT_ESTIMATE],
        std_mean=np.sqrt(df[VARIANCE] / df[denominator]),
        alpha=1 - interval_size,
        alternative=TWO_SIDED,
    )
    return df


def p_value(df: DataFrame, **kwargs: Dict[str, str]) -> Series:
    _, p_value = _zstat_generic(
        value1=df[POINT_ESTIMATE + SFX2],
        value2=df[POINT_ESTIMATE + SFX1],
        std_diff=df[STD_ERR],
        alternative=df[PREFERENCE_TEST].values[0],
        diff=df[NULL_HYPOTHESIS],
    )
    return p_value


def ci(df: DataFrame, alpha_column: str, **kwargs: Dict[str, str]) -> Tuple[Series, Series]:
    return _zconfint_generic(
        mean=df[DIFFERENCE], std_mean=df[STD_ERR], alpha=df[alpha_column], alternative=df[PREFERENCE_TEST].values[0]
    )


def achieved_power(df: DataFrame, mde: float, alpha: float, **kwargs: Dict[str, str]) -> DataFrame:
    denominator = kwargs[DENOMINATOR]
    v1, v2 = df[VARIANCE + SFX1], df[VARIANCE + SFX2]
    n1, n2 = df[denominator + SFX1], df[denominator + SFX2]

    var_pooled = ((n1 - 1) * v1 + (n2 - 1) * v2) / (n1 + n2 - 2)

    return power_calculation(mde, var_pooled, alpha, n1, n2)


def compute_sequential_adjusted_alpha(df: DataFrame, **kwargs: Dict[str, str]):
    denominator = kwargs[DENOMINATOR]
    final_expected_sample_size_column = kwargs[FINAL_EXPECTED_SAMPLE_SIZE]
    ordinal_group_column = kwargs[ORDINAL_GROUP_COLUMN]
    n_comparisons = kwargs[NUMBER_OF_COMPARISONS]

    if not df.reset_index()[ordinal_group_column].is_unique:
        raise ValueError("Ordinal values cannot be duplicated")

    def adjusted_alphas_for_group(grp: DataFrame) -> Series:
        return (
            sequential_bounds(
                t=grp["sample_size_proportions"].values,
                alpha=grp[ALPHA].values[0] / n_comparisons,
                sides=2 if (grp[PREFERENCE_TEST] == TWO_SIDED).all() else 1,
            )
            .df.set_index(grp.index)
            .assign(
                **{
                    ADJUSTED_ALPHA: lambda df: df.apply(
                        lambda row: 2 * (1 - st.norm.cdf(row["zb"]))
                        if (grp[PREFERENCE_TEST] == TWO_SIDED).all()
                        else 1 - st.norm.cdf(row["zb"]),
                        axis=1,
                    )
                }
            )
        )[["zb", ADJUSTED_ALPHA]]

    comparison_total_column = "comparison_total_" + denominator
    return Series(
        data=(
            df.assign(**{comparison_total_column: df[denominator + SFX1] + df[denominator + SFX2]})
            .assign(
                max_sample_size=lambda df: df[[comparison_total_column, final_expected_sample_size_column]]
                .max(axis=1)
                .max()
            )
            .assign(sample_size_proportions=lambda df: df[comparison_total_column] / df["max_sample_size"])
            .pipe(adjusted_alphas_for_group)[ADJUSTED_ALPHA]
        ),
        name=ADJUSTED_ALPHA,
    )


def ci_for_multiple_comparison_methods(
    df: DataFrame,
    correction_method: str,
    alpha: float,
    w: float = 1.0,
) -> Tuple[Union[Series, float], Union[Series, float]]:
    if TWO_SIDED in df[PREFERENCE_TEST]:
        raise ValueError(
            "CIs can only be produced for one-sided tests when other multiple test corrections "
            "methods than bonferroni are applied"
        )
    m_scal = len(df)
    num_significant = sum(df[IS_SIGNIFICANT])
    r = m_scal - num_significant

    def _aw(W: float, alpha: float, m_scal: float, r: int):
        return alpha * (1 - (1 - W) * (m_scal - r) / m_scal)

    def _bw(W: float, alpha: float, m_scal: float, r: int):
        return 1 - (1 - alpha) / np.power((1 - (1 - W) * (1 - np.power((1 - alpha), (1 / m_scal)))), (m_scal - r))

    if correction_method in [HOLM, SPOT_1_HOLM]:
        adjusted_alpha_rej_equal_m = 1 - alpha / m_scal
        adjusted_alpha_rej_less_m = 1 - (1 - w) * (alpha / m_scal)
        adjusted_alpha_accept = 1 - _aw(w, alpha, m_scal, r) / r if r != 0 else 0
    elif correction_method in [HOMMEL, SIMES_HOCHBERG, SPOT_1_HOMMEL, SPOT_1_SIMES_HOCHBERG]:
        adjusted_alpha_rej_equal_m = np.power((1 - alpha), (1 / m_scal))
        adjusted_alpha_rej_less_m = 1 - (1 - w) * (1 - np.power((1 - alpha), (1 / m_scal)))
        adjusted_alpha_accept = 1 - _bw(w, alpha, m_scal, r) / r if r != 0 else 0
    else:
        raise ValueError(
            "CIs not supported for correction method. "
            f"Supported methods: {HOMMEL}, {HOLM}, {SIMES_HOCHBERG},"
            f"{SPOT_1_HOLM}, {SPOT_1_HOMMEL} and {SPOT_1_SIMES_HOCHBERG}"
        )

    def _compute_ci_for_row(row: Series) -> Tuple[float, float]:
        if row[IS_SIGNIFICANT] and num_significant == m_scal:
            alpha_adj = adjusted_alpha_rej_equal_m
        elif row[IS_SIGNIFICANT] and num_significant < m_scal:
            alpha_adj = adjusted_alpha_rej_less_m
        else:
            alpha_adj = adjusted_alpha_accept

        ci_sign = -1 if row[PREFERENCE_TEST] == "larger" else 1
        bound1 = row[DIFFERENCE] + ci_sign * st.norm.ppf(alpha_adj) * row[STD_ERR]
        if ci_sign == -1:
            bound2 = max(row[NULL_HYPOTHESIS], bound1)
        else:
            bound2 = min(row[NULL_HYPOTHESIS], bound1)

        bound = bound2 if row[IS_SIGNIFICANT] else bound1

        lower = bound if row[PREFERENCE_TEST] == "larger" else -np.inf
        upper = bound if row[PREFERENCE_TEST] == "smaller" else np.inf

        row[ADJUSTED_LOWER] = lower
        row[ADJUSTED_UPPER] = upper

        return row

    ci_df = df.apply(_compute_ci_for_row, axis=1)[[ADJUSTED_LOWER, ADJUSTED_UPPER]]

    return ci_df[ADJUSTED_LOWER], ci_df[ADJUSTED_UPPER]


def ci_width(
    z_alpha, binary, non_inferiority, hypothetical_effect, control_avg, control_var, control_count, treatment_count
) -> Union[Series, float]:
    treatment_var = _get_hypothetical_treatment_var(
        binary, non_inferiority, control_avg, control_var, hypothetical_effect
    )
    _, std_err = _unequal_var_ttest_denom(control_var, control_count, treatment_var, treatment_count)

    return 2 * z_alpha * std_err


def powered_effect(
    df: DataFrame,
    z_alpha: float,
    z_power: float,
    binary: bool,
    non_inferiority: bool,
    avg_column: float,
    var_column: float,
) -> Series:
    if binary and not non_inferiority:
        effect = df.apply(
            lambda row: _search_MDE_binary_local_search(
                control_avg=row[avg_column],
                control_var=row[var_column],
                non_inferiority=False,
                kappa=row["kappa"],
                proportion_of_total=row["proportion_of_total"],
                current_number_of_units=row["current_number_of_units"],
                z_alpha=z_alpha,
                z_power=z_power,
            )[0],
            axis=1,
        )
    else:
        treatment_var = _get_hypothetical_treatment_var(
            binary_metric=binary,
            non_inferiority=df[NIM] is not None,
            control_avg=df[avg_column],
            control_var=df[var_column],
            hypothetical_effect=0,
        )
        n2_partial = np.power((z_alpha + z_power), 2) * (df[var_column] / df["kappa"] + treatment_var)
        effect = np.sqrt(
            (1 / (df["current_number_of_units"] * df["proportion_of_total"])) * (n2_partial + df["kappa"] * n2_partial)
        )

    return effect


def required_sample_size(
    binary: Union[Series, bool],
    non_inferiority: Union[Series, bool],
    hypothetical_effect: Union[Series, float],
    control_avg: Union[Series, float],
    control_var: Union[Series, float],
    z_alpha: float = None,
    kappa: float = None,
    proportion_of_total: Union[Series, float] = None,
    z_power: float = None,
) -> Union[Series, float]:
    if kappa is None:
        raise ValueError("kappa is None, must be postive float")
    if proportion_of_total is None:
        raise ValueError("proportion_of_total is None, must be between 0 and 1")

    treatment_var = np.vectorize(_get_hypothetical_treatment_var)(
        binary, non_inferiority, control_avg, control_var, hypothetical_effect
    )

    n2 = _treatment_group_sample_size(
        z_alpha=z_alpha,
        z_power=z_power,
        hypothetical_effect=hypothetical_effect,
        control_var=control_var,
        treatment_var=treatment_var,
        kappa=kappa,
    )
    required_sample_size = np.ceil((n2 + n2 * kappa) / proportion_of_total)
    return required_sample_size


def _search_MDE_binary_local_search(
    control_avg: float,
    control_var: float,
    non_inferiority: bool,
    kappa: float,
    proportion_of_total: float,
    current_number_of_units: float,
    z_alpha: float = None,
    z_power: float = None,
):
    def f(x):
        return _find_current_powered_effect(
            hypothetical_effect=x,
            control_avg=control_avg,
            control_var=control_var,
            binary=True,
            non_inferiority=non_inferiority,
            kappa=kappa,
            proportion_of_total=proportion_of_total,
            current_number_of_units=current_number_of_units,
            z_alpha=z_alpha,
            z_power=z_power,
        )

    max_val = 1 - control_avg
    min_val = min(10e-9, max_val)

    if min_val == max_val:
        # corner case that crashes the optimizer
        return min_val, f(min_val)

    max_iter = 100  # max number of iterations before falling back to slow grid search

    # we stop immediately if a solution was found that is "good enough". A threshold of
    # 1 indicates that
    # the approximated number of units (based on the current effect candidate) is off by
    # at most 1.0
    goodness_threshold = 1.0

    curr_iter = 0
    best_x = None
    best_fun = float("inf")

    bounds_queue = [(min_val, max_val)]

    while curr_iter < max_iter and best_fun > goodness_threshold:
        # take next value from queue
        interval = bounds_queue.pop(0)

        # conduct a bounded local search, using a very small tol value improved
        # performance during tests
        # result = optimize.minimize_scalar(f, bounds=(interval[0], interval[1]),
        # method='bounded', tol=10e-14)
        result = optimize.minimize_scalar(
            f, bounds=(interval[0], interval[1]), method="bounded", options={"xatol": 10e-14, "maxiter": 50}
        )

        if result.fun < best_fun:
            best_x = result.x
            best_fun = result.fun

        curr_iter += 1

        # add new bounds to the queue
        interval_split = (interval[0] + interval[1]) / 2
        bounds_queue.append((interval[0], interval_split))
        bounds_queue.append((interval_split, interval[1]))

    if best_fun <= goodness_threshold:
        return best_x, best_fun
    else:  # check if grid search finds a better solution
        alt_result_x, alt_result_fun = _search_MDE_binary(
            control_avg,
            control_var,
            non_inferiority,
            kappa,
            proportion_of_total,
            current_number_of_units,
            z_alpha,
            z_power,
            return_cost_val=True,
        )

        return (alt_result_x, alt_result_fun) if alt_result_fun < best_fun else (best_x, best_fun)


def _search_MDE_binary(
    control_avg: float,
    control_var: float,
    non_inferiority: bool,
    kappa: float,
    proportion_of_total: float,
    current_number_of_units: float,
    z_alpha: float = None,
    z_power: float = None,
    return_cost_val=False,
):
    candidate_effects = np.linspace(10e-9, 1 - control_avg, num=2000)
    for i in range(2):
        test = []
        for effect in candidate_effects:
            test.append(
                _find_current_powered_effect(
                    hypothetical_effect=effect,
                    control_avg=control_avg,
                    control_var=control_var,
                    binary=True,
                    non_inferiority=non_inferiority,
                    kappa=kappa,
                    proportion_of_total=proportion_of_total,
                    current_number_of_units=current_number_of_units,
                    z_alpha=z_alpha,
                    z_power=z_power,
                )
            )

        test = np.array(test)
        index = [idx for idx, element in enumerate(test) if element == test.min()]
        if len(index) != 1:
            index = [index[int(np.ceil(len(index) / 2))]]
        if i == 0:
            if index[0] == 9999:
                return np.inf
            lower_effect_bound = 10e-9 if index[0] == 0 else candidate_effects[index[0] - 1]
            candidate_effects = np.linspace(lower_effect_bound, candidate_effects[index[0]], num=10000)

    index = [idx for idx, element in enumerate(test) if element == test.min()]

    return candidate_effects[index[0]], test[index[0]] if return_cost_val else candidate_effects[index[0]]


def _treatment_group_sample_size(
    z_alpha: float,
    z_power: float,
    hypothetical_effect: float,
    control_var: float,
    treatment_var: float,
    kappa: float,
) -> float:
    return np.ceil(np.power((z_alpha + z_power) / abs(hypothetical_effect), 2) * (control_var / kappa + treatment_var))


def _find_current_powered_effect(
    hypothetical_effect: float,
    control_avg: float,
    control_var: float,
    binary: bool,
    non_inferiority: bool,
    kappa: float,
    proportion_of_total: float,
    current_number_of_units: float,
    z_power: float = None,
    z_alpha: float = None,
) -> float:
    treatment_var = _get_hypothetical_treatment_var(
        binary_metric=binary,
        non_inferiority=non_inferiority,
        control_avg=control_avg,
        control_var=control_var,
        hypothetical_effect=hypothetical_effect,
    )
    n2 = _treatment_group_sample_size(
        z_alpha,
        z_power,
        hypothetical_effect,
        control_var,
        treatment_var,
        kappa,
    )

    return np.power(current_number_of_units - ((n2 + n2 * kappa) / proportion_of_total), 2)


def _get_hypothetical_treatment_var(
    binary_metric: bool,
    non_inferiority: bool,
    control_avg: float,
    control_var: float,
    hypothetical_effect: float,
) -> float:
    if binary_metric and not non_inferiority:
        # For binary metrics, the variance can be derived from the average. However,
        # we do *not* do this for
        # non-inferiority tests because for non-inferiority tests, the basic assumption
        # is that the
        # mean of the control group and treatment group are identical.
        return (control_avg + hypothetical_effect) * (1 - (control_avg + hypothetical_effect))
    else:
        return control_var



# Copyright 2017-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Union, Iterable, List, Tuple

from pandas import DataFrame

from ..constants import NIM_TYPE


class ConfidenceComputerABC(ABC):
    @abstractmethod
    def compute_summary(self, verbose: bool) -> DataFrame:
        """Return Pandas DataFrame with summary statistics."""
        pass

    @abstractmethod
    def compute_difference(
        self,
        level_1: Union[str, Iterable],
        level_2: Union[str, Iterable],
        absolute: bool,
        groupby: Union[str, Iterable],
        nims: NIM_TYPE,
        final_expected_sample_size_column: str,
        verbose: bool,
        mde_column: str,
    ) -> DataFrame:
        """Return dataframe containing the difference in means between
        group 1 and 2, p-value and confidence interval
        """
        pass

    @abstractmethod
    def compute_multiple_difference(
        self,
        level: Union[str, Iterable],
        absolute: bool,
        groupby: Union[str, Iterable],
        level_as_reference: bool,
        nims: NIM_TYPE,
        final_expected_sample_size_column: str,
        verbose: bool,
        mde_column: str,
    ) -> DataFrame:
        """Return dataframe containing the difference in means between
        level and all other groups, with p-value and confidence interval
        """
        pass

    def compute_differences(
        self,
        levels: List[Tuple],
        absolute: bool,
        groupby: Union[str, Iterable],
        nims: NIM_TYPE,
        final_expected_sample_size_column: str,
        verbose: bool,
        mde_column: str,
    ) -> DataFrame:
        """Return dataframe containing the difference in means between
        level and all other groups, with p-value and confidence interval
        """
        pass

    def achieved_power(
        self,
        level_1: Union[str, Iterable],
        level_2: Union[str, Iterable],
        mde: float,
        alpha: float,
        groupby: Union[str, Iterable],
    ) -> DataFrame:
        """Calculated the achieved power of test of differences between
        level 1 and level 2 given a targeted MDE.

        Args:
            level_1 (str, tuple of str): Name of first level.
            level_2 (str, tuple of str): Name of second level.
            mde (float): Absolute minimal detectable effect size.
            alpha (float): Type I error rate, cutoff value for determining
                statistical significance.
            groupby (str): Name of column.
                If specified, will return the difference for each level
                of the grouped dimension.

        Returns:
            Pandas DataFrame with the following columns:
            - level_1: Name of level 1.
            - level_2: Name of level 2.
            - power: 1 - B, where B is the likelihood of a Type II (false
                negative) error.
        """
        pass



import numpy as np
import pandas as pd

from spotify_confidence.analysis.constants import (
    SPOT_1,
    REQUIRED_SAMPLE_SIZE_METRIC,
    CI_WIDTH,
    POWERED_EFFECT,
    POINT_ESTIMATE,
    ADJUSTED_LOWER,
)
from spotify_confidence.analysis.frequentist.sample_size_calculator import SampleSizeCalculator
from spotify_confidence.analysis.frequentist.z_test import ZTest

pd.options.display.max_columns = None
pd.options.display.width = 1000


class TestSampleSizeCalculator(object):
    def test_sample_size_1(self):
        df = pd.DataFrame(
            columns=["metric_name", "binary", "avg", "var", "mde", "nim", "preference"],
            data=[
                ["share_bananas_1d", True, 0.7, 0.21, 0.00617, None, "increase"],
                ["bananas_per_user_7d", False, 4.56, 2.13, 0.01, None, "increase"],
            ],
        )

        ssc = SampleSizeCalculator(
            data_frame=df,
            point_estimate_column="avg",
            var_column="var",
            metric_column="metric_name",
            is_binary_column="binary",
            interval_size=0.99,
            power=0.8,
            correction_method=SPOT_1,
        )
        treatment_weights = [5000, 2000, 3000]
        ss = ssc.sample_size(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
        )

        assert len(ss) == len(df)
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0] / 1042868 < 1.001
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[1] / 95459 < 1.001
        assert ss[CI_WIDTH].isna().all()

        optimal_weights, optimal_sample_size = ssc.optimal_weights_and_sample_size(ss, len(treatment_weights))
        assert len(optimal_weights) == len(treatment_weights)
        assert 0.999 < optimal_sample_size / 894863 < 1.001

        powered_effect_df = ssc.powered_effect(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
            sample_size=ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0],
        )

        relative_powered_effect = powered_effect_df[POWERED_EFFECT] / powered_effect_df["avg"]

        assert np.isclose(relative_powered_effect.values[0], df["mde"].values[0], rtol=1e-3)

    def test_sample_size_2(self):
        df = pd.DataFrame(
            columns=["metric_name", "binary", "avg", "var", "mde", "nim", "preference"],
            data=[
                ["share_bananas_1d", True, 0.7, 0.21, 0.00617, None, None],
                ["bananas_per_user_7d", False, 4.56, 2.13, 0.01, None, "increase"],
            ],
        )

        ssc = SampleSizeCalculator(
            data_frame=df,
            point_estimate_column="avg",
            var_column="var",
            metric_column="metric_name",
            is_binary_column="binary",
            interval_size=0.99,
            power=0.8,
            correction_method=SPOT_1,
        )
        treatment_weights = [5000, 2000, 3000]
        ss = ssc.sample_size(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
        )

        assert len(ss) == len(df)
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0] / 1170185 < 1.001
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[1] / 95459 < 1.001
        assert ss[CI_WIDTH].isna().all()

        optimal_weights, optimal_sample_size = ssc.optimal_weights_and_sample_size(ss, len(treatment_weights))
        assert len(optimal_weights) == len(treatment_weights)
        assert 0.999 < optimal_sample_size / 1004113 < 1.001

        powered_effect_df = ssc.powered_effect(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
            sample_size=ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0],
        )

        relative_powered_effect = powered_effect_df[POWERED_EFFECT] / powered_effect_df["avg"]

        assert np.isclose(relative_powered_effect.values[0], df["mde"].values[0], rtol=1e-3)

    def test_sample_size_3(self):
        df = pd.DataFrame(
            columns=["metric_name", "binary", "avg", "var", "mde", "nim", "preference"],
            data=[
                ["share_bananas_1d", True, 0.7, 0.21, 0.00617, None, None],
                ["bananas_per_user_7d", False, 4.56, 2.13, 0.01, None, "increase"],
            ],
        )

        ssc = SampleSizeCalculator(
            data_frame=df,
            point_estimate_column="avg",
            var_column="var",
            metric_column="metric_name",
            is_binary_column="binary",
            interval_size=0.99,
            power=0.8,
            correction_method=SPOT_1,
        )
        treatment_weights = [5000, 2000]
        ss = ssc.sample_size(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
        )

        assert len(ss) == len(df)
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0] / 730009 < 1.001
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[1] / 58621 < 1.001
        assert ss[CI_WIDTH].isna().all()

        optimal_weights, optimal_sample_size = ssc.optimal_weights_and_sample_size(ss, len(treatment_weights))
        assert len(optimal_weights) == len(treatment_weights)
        assert 0.999 < optimal_sample_size / 596991 < 1.001

        powered_effect_df = ssc.powered_effect(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
            sample_size=ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0],
        )

        relative_powered_effect = powered_effect_df[POWERED_EFFECT] / powered_effect_df["avg"]

        assert np.isclose(relative_powered_effect.values[0], df["mde"].values[0], rtol=1e-3)

    def test_sample_size_4(self):
        df = pd.DataFrame(
            columns=["metric_name", "binary", "avg", "var", "mde", "nim", "preference"],
            data=[
                ["share_bananas_1d", True, 0.7, 0.21, None, 0.00617, "increase"],
                ["bananas_per_user_7d", False, 4.56, 2.13, 0.01, None, "increase"],
            ],
        )

        ssc = SampleSizeCalculator(
            data_frame=df,
            point_estimate_column="avg",
            var_column="var",
            metric_column="metric_name",
            is_binary_column="binary",
            interval_size=0.99,
            power=0.8,
            correction_method=SPOT_1,
        )
        treatment_weights = [5000, 2000]
        ss = ssc.sample_size(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
        )

        assert len(ss) == len(df)
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0] / 718056 < 1.001
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[1] / 65337 < 1.001
        assert ss[CI_WIDTH].isna().all()

        optimal_weights, optimal_sample_size = ssc.optimal_weights_and_sample_size(ss, len(treatment_weights))
        assert len(optimal_weights) == len(treatment_weights)
        assert 0.999 < optimal_sample_size / 586168 < 1.001

        powered_effect_df = ssc.powered_effect(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
            sample_size=ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0],
        )

        relative_powered_effect = powered_effect_df[POWERED_EFFECT] / powered_effect_df["avg"]

        assert np.isclose(relative_powered_effect.values[0], df["nim"].values[0], rtol=1e-3)

    def test_sample_size_5(self):
        df = pd.DataFrame(
            columns=["metric_name", "binary", "avg", "var", "mde", "nim", "preference"],
            data=[
                ["share_bananas_1d", True, 0.7, 0.21, None, 0.00617, "increase"],
            ],
        )

        ssc = SampleSizeCalculator(
            data_frame=df,
            point_estimate_column="avg",
            var_column="var",
            metric_column="metric_name",
            is_binary_column="binary",
            interval_size=0.99,
            power=0.8,
            correction_method=SPOT_1,
        )
        treatment_weights = [5000, 2000]
        ss = ssc.sample_size(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
        )

        assert len(ss) == len(df)
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0] / 553620 < 1.001
        assert ss[CI_WIDTH].isna().all()

        optimal_weights, optimal_sample_size = ssc.optimal_weights_and_sample_size(ss, len(treatment_weights))
        assert len(optimal_weights) == len(treatment_weights)
        assert 0.999 < optimal_sample_size / 451934 < 1.001

        powered_effect_df = ssc.powered_effect(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
            sample_size=ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0],
        )

        relative_powered_effect = powered_effect_df[POWERED_EFFECT] / powered_effect_df["avg"]

        assert np.isclose(relative_powered_effect.values[0], df["nim"].values[0], rtol=1e-3)

    def test_sample_size_6(self):
        df = pd.DataFrame(
            columns=["metric_name", "binary", "avg", "var", "mde", "nim", "preference"],
            data=[
                ["bananas_per_user_7d", False, 4.56, 2.13, 0.01, None, None],
            ],
        )

        ssc = SampleSizeCalculator(
            data_frame=df,
            point_estimate_column="avg",
            var_column="var",
            metric_column="metric_name",
            is_binary_column="binary",
            interval_size=0.99,
            power=0.8,
            correction_method=SPOT_1,
        )
        treatment_weights = [5000, 2000]
        ss = ssc.sample_size(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
        )

        assert len(ss) == len(df)
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0] / 58622 < 1.001
        assert ss[CI_WIDTH].isna().all()

        optimal_weights, optimal_sample_size = ssc.optimal_weights_and_sample_size(ss, len(treatment_weights))
        assert len(optimal_weights) == len(treatment_weights)
        assert 0.999 < optimal_sample_size / 47854 < 1.001

        powered_effect_df = ssc.powered_effect(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
            sample_size=ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0],
        )

        relative_powered_effect = powered_effect_df[POWERED_EFFECT] / powered_effect_df["avg"]

        assert np.isclose(relative_powered_effect.values[0], df["mde"].values[0], rtol=1e-3)

    def test_sample_size_7(self):
        df = pd.DataFrame(
            columns=["metric_name", "binary", "avg", "var", "mde", "nim", "preference"],
            data=[
                ["share_bananas_1d", True, 0.7, 0.21, 0.00617, None, None],
                ["bananas_per_user_7d", False, 4.56, 2.13, 0.01, None, "increase"],
            ],
        )

        ssc = SampleSizeCalculator(
            data_frame=df,
            point_estimate_column="avg",
            var_column="var",
            metric_column="metric_name",
            is_binary_column="binary",
            interval_size=0.95,
            power=0.9,
            correction_method=SPOT_1,
        )
        treatment_weights = [5000, 2000]
        ss = ssc.sample_size(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
        )

        assert len(ss) == len(df)
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0] / 680575 < 1.001
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[1] / 52741 < 1.001
        assert ss[CI_WIDTH].isna().all()

        optimal_weights, optimal_sample_size = ssc.optimal_weights_and_sample_size(ss, len(treatment_weights))
        assert len(optimal_weights) == len(treatment_weights)
        assert 0.999 < optimal_sample_size / 556565 < 1.001

        powered_effect_df = ssc.powered_effect(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
            sample_size=ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0],
        )

        relative_powered_effect = powered_effect_df[POWERED_EFFECT] / powered_effect_df["avg"]

        assert np.isclose(relative_powered_effect.values[0], df["mde"].values[0], rtol=1e-3)

    def test_sample_size_8(self):
        df = pd.DataFrame(
            columns=["metric_name", "country", "binary", "avg", "var", "mde", "nim", "preference"],
            data=[
                ["share_bananas_1d", "denmark", True, 0.7, 0.21, 0.01, None, None],
                ["share_bananas_1d", "sweden", True, 0.4, 0.24, 0.01, None, None],
                ["bananas_per_user_7d", "denmark", False, 4.56, 2.13, 0.01, None, None],
                ["bananas_per_user_7d", "sweden", False, 3.81, 7.11, 0.01, None, None],
            ],
        )

        ssc = SampleSizeCalculator(
            data_frame=df,
            point_estimate_column="avg",
            var_column="var",
            metric_column="metric_name",
            categorical_group_columns="country",
            is_binary_column="binary",
            interval_size=0.99,
            power=0.8,
            correction_method=SPOT_1,
        )
        treatment_weights = [1, 1, 1, 1]
        ss = ssc.sample_size(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
        )

        assert len(ss) == len(df)
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0] / 595876 < 1.001
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[1] / 2103232 < 1.001
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[2] / 143396 < 1.001
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[3] / 685656 < 1.001
        assert ss[CI_WIDTH].isna().all()

        powered_effect_df = ssc.powered_effect(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
            sample_size=ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0],
        )

        relative_powered_effect = powered_effect_df[POWERED_EFFECT] / powered_effect_df["avg"]

        assert np.isclose(relative_powered_effect.values[0], df["mde"].values[0], rtol=1e-3)

    def test_sample_size_calculation_ciwidth_nimless_with_expected_sample_size(self):
        df = pd.DataFrame(
            columns=["metric_name", "binary", "avg", "var", "mde", "nim", "preference", "expected_sample_size"],
            data=[
                ["share_bananas_1d", True, 0.7, 0.21, None, 0.0, "increase", 1e6],
                ["bananas_per_user_7d", False, 4.56, 2.13, 0.01, None, "increase", int(1e6)],
            ],
        )

        ssc = SampleSizeCalculator(
            data_frame=df,
            point_estimate_column="avg",
            var_column="var",
            metric_column="metric_name",
            is_binary_column="binary",
            interval_size=0.95,
            power=0.8,
            correction_method=SPOT_1,
        )
        treatment_weights = [5000, 2000, 3000]
        ss = ssc.sample_size(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
            final_expected_sample_size_column="expected_sample_size",
        )

        assert len(ss) == len(df)
        assert ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0] == float("inf")
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[1] / 75345 < 1.001

        np.testing.assert_almost_equal(ss[CI_WIDTH].values[0], 0.0047527)
        np.testing.assert_almost_equal(ss[CI_WIDTH].values[1], 0.0151362)

        powered_effect_df = ssc.powered_effect(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
            sample_size=ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0],
        )

        relative_powered_effect = powered_effect_df[POWERED_EFFECT] / powered_effect_df["avg"]

        assert np.isnan(relative_powered_effect.values[0])

    def test_sample_size_calculation_ciwidth_matches_real_width_returned_by_onesided_test(self):
        df = pd.DataFrame(
            columns=["metric_name", "binary", "avg", "var", "mde", "nim", "preference", "expected_sample_size"],
            data=[
                ["bananas_per_user_7d", False, 4.56, 2.13, 0.01, None, "increase", int(1e6)],
            ],
        )

        ssc = SampleSizeCalculator(
            data_frame=df,
            point_estimate_column="avg",
            var_column="var",
            metric_column="metric_name",
            is_binary_column="binary",
            interval_size=0.95,
            power=0.8,
            correction_method=SPOT_1,
        )
        treatment_weights = [5000, 5000]
        ss = ssc.sample_size(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
            final_expected_sample_size_column="expected_sample_size",
        )

        assert len(ss) == len(df)
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0] / 25334 < 1.001
        np.testing.assert_almost_equal(ss[CI_WIDTH].values[0], 0.0096023)

        powered_effect_df = ssc.powered_effect(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
            sample_size=ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0],
        )

        relative_powered_effect = powered_effect_df[POWERED_EFFECT] / powered_effect_df["avg"]

        assert np.isclose(relative_powered_effect.values[0], df["mde"].values[0], rtol=1e-3)

    def test_sample_size_calculation_ciwidth_matches_real_width_returned_by_twosided_test_with_direction(self):
        df = pd.DataFrame(
            columns=["metric_name", "binary", "avg", "var", "mde", "nim", "preference", "expected_sample_size"],
            data=[
                ["bananas_per_user_7d", False, 4.56, 2.13, 0.01, None, None, int(1e6)],
            ],
        )

        ssc = SampleSizeCalculator(
            data_frame=df,
            point_estimate_column="avg",
            var_column="var",
            metric_column="metric_name",
            is_binary_column="binary",
            interval_size=0.95,
            power=0.8,
            correction_method=SPOT_1,
        )
        treatment_weights = [5000, 5000]
        ss = ssc.sample_size(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
            final_expected_sample_size_column="expected_sample_size",
        )

        assert len(ss) == len(df)
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0] / 32162 < 1.001
        np.testing.assert_almost_equal(ss[CI_WIDTH].values[0], 0.01144189)

        powered_effect_df = ssc.powered_effect(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
            sample_size=ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0],
        )

        relative_powered_effect = powered_effect_df[POWERED_EFFECT] / powered_effect_df["avg"]

        assert np.isclose(relative_powered_effect.values[0], df["mde"].values[0], rtol=1e-3)

    def test_ci_bound_equal_nim(self):
        mde_pp = 0.15
        baseline = 0.48
        df = pd.DataFrame(
            columns=["metric_name", "binary", "avg", "var", "mde", "nim", "preference"],
            data=[
                [
                    "share_users_with_bananas",
                    False,
                    baseline,
                    baseline * (1 - baseline),
                    None,
                    mde_pp / 100 / baseline,
                    "increase",
                ],
            ],
        )

        ssc = SampleSizeCalculator(
            data_frame=df,
            point_estimate_column="avg",
            var_column="var",
            metric_column="metric_name",
            is_binary_column="binary",
            interval_size=0.95,
            power=0.5,
            correction_method="bonferroni",
        )
        treatment_weights = [0.5, 0.125, 0.125, 0.125, 0.125]
        ss = ssc.sample_size(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
        )

        assert len(ss) == len(df)
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0] / 5573168 < 1.001

        powered_effect_df = ssc.powered_effect(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
            sample_size=ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0],
        )

        relative_powered_effect = powered_effect_df[POWERED_EFFECT] / powered_effect_df["avg"]

        assert np.isclose(relative_powered_effect.values[0], df["nim"].values[0], rtol=1e-3)

        df_test = pd.DataFrame(
            data={
                "test_group": [
                    "control",
                    "same as control",
                    "MDE above control",
                    "2% of MDE above control",
                    "MDE below control",
                ],
                "num_users": [2786584, 696646, 696646, 696646, 696646],
                "num_users_with_bananas": [1337560.32, 334390.08, 342000, 341077.8816, 340000],
            }
        )

        test = ZTest(
            data_frame=df_test,
            numerator_column="num_users_with_bananas",
            numerator_sum_squares_column="num_users_with_bananas",
            denominator_column="num_users",
            categorical_group_columns="test_group",
            interval_size=0.95,
            power=0.5,
        )

        summary = test.summary()
        assert np.isclose(summary[POINT_ESTIMATE].values[0], 0.48)
        assert np.isclose(summary[POINT_ESTIMATE].values[1], 0.48)

        diff = test.multiple_difference(
            level="control", level_as_reference=True, non_inferiority_margins=(mde_pp / 100 / baseline, "increase")
        )

        assert np.isclose(diff.query("level_2 == 'same as control'")[ADJUSTED_LOWER], -mde_pp / 100)

    def test_ci_bound_equal_nim_2(self):
        mde_pp = 0.15
        baseline = 0.48
        df = pd.DataFrame(
            columns=["metric_name", "binary", "avg", "var", "mde", "nim", "preference"],
            data=[
                [
                    "share_users_with_bananas",
                    False,
                    baseline,
                    baseline * (1 - baseline),
                    None,
                    mde_pp / 100 / baseline,
                    "increase",
                ],
            ],
        )

        ssc = SampleSizeCalculator(
            data_frame=df,
            point_estimate_column="avg",
            var_column="var",
            metric_column="metric_name",
            is_binary_column="binary",
            interval_size=0.95,
            power=0.5,
            correction_method="bonferroni",
        )
        treatment_weights = [0.75, 0.25]
        ss = ssc.sample_size(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
        )

        required_sample_size = ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0]

        powered_effect_df = ssc.powered_effect(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
            sample_size=ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0],
        )

        relative_powered_effect = powered_effect_df[POWERED_EFFECT] / powered_effect_df["avg"]

        assert np.isclose(relative_powered_effect.values[0], df["nim"].values[0], rtol=1e-3)

        df_test = pd.DataFrame(
            data={
                "test_group": [
                    "control",
                    "same as control",
                ],
                "num_users": [
                    required_sample_size * treatment_weights[0],
                    required_sample_size * treatment_weights[1],
                ],
                "num_users_with_bananas": [
                    required_sample_size * treatment_weights[0] * baseline,
                    required_sample_size * treatment_weights[1] * baseline,
                ],
            }
        )

        test = ZTest(
            data_frame=df_test,
            numerator_column="num_users_with_bananas",
            numerator_sum_squares_column="num_users_with_bananas",
            denominator_column="num_users",
            categorical_group_columns="test_group",
            interval_size=0.95,
            power=0.5,
        )

        summary = test.summary()
        assert np.isclose(summary[POINT_ESTIMATE].values[0], 0.48)
        assert np.isclose(summary[POINT_ESTIMATE].values[1], 0.48)

        diff = test.multiple_difference(
            level="control", level_as_reference=True, non_inferiority_margins=(mde_pp / 100 / baseline, "increase")
        )

        assert np.isclose(diff[ADJUSTED_LOWER], -mde_pp / 100)

    def test_ci_bound_equal_mde(self):
        mde_pp = 0.15
        baseline = 0.48
        df = pd.DataFrame(
            columns=["metric_name", "binary", "avg", "var", "mde", "nim", "preference"],
            data=[
                [
                    "share_users_with_bananas",
                    True,
                    baseline,
                    baseline * (1 - baseline),
                    mde_pp / 100 / baseline,
                    None,
                    None,
                ],
            ],
        )

        ssc = SampleSizeCalculator(
            data_frame=df,
            point_estimate_column="avg",
            var_column="var",
            metric_column="metric_name",
            is_binary_column="binary",
            interval_size=0.95,
            power=0.5,
            correction_method="bonferroni",
        )
        treatment_weights = [0.75, 0.25]
        ss = ssc.sample_size(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
        )

        required_sample_size = ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0]

        powered_effect_df = ssc.powered_effect(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
            sample_size=ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0],
        )

        relative_powered_effect = powered_effect_df[POWERED_EFFECT] / powered_effect_df["avg"]

        assert np.isclose(relative_powered_effect.values[0], df["mde"].values[0], rtol=1e-3)

        df_test = pd.DataFrame(
            data={
                "test_group": [
                    "control",
                    "same as control",
                ],
                "num_users": [
                    required_sample_size * treatment_weights[0],
                    required_sample_size * treatment_weights[1],
                ],
                "num_users_with_bananas": [
                    required_sample_size * treatment_weights[0] * baseline,
                    required_sample_size * treatment_weights[1] * baseline,
                ],
            }
        )

        test = ZTest(
            data_frame=df_test,
            numerator_column="num_users_with_bananas",
            numerator_sum_squares_column="num_users_with_bananas",
            denominator_column="num_users",
            categorical_group_columns="test_group",
            interval_size=0.95,
            power=0.5,
        )

        summary = test.summary()
        assert np.isclose(summary[POINT_ESTIMATE].values[0], 0.48)
        assert np.isclose(summary[POINT_ESTIMATE].values[1], 0.48)

        diff = test.multiple_difference(
            level="control",
            level_as_reference=True,  # non_inferiority_margins=(mde_pp / 100 / baseline, "increase")
        )

        assert np.isclose(diff[ADJUSTED_LOWER], -mde_pp / 100, rtol=0.001)

    def test_sample_size_with_nan(self):
        df = pd.DataFrame(
            columns=["metric_name", "binary", "avg", "var", "mde", "nim", "preference"],
            data=[
                ["share_bananas_1d", True, None, None, 0.00617, None, "increase"],
                ["bananas_per_user_7d", False, 4.56, 2.13, 0.01, None, "increase"],
            ],
        )

        ssc = SampleSizeCalculator(
            data_frame=df,
            point_estimate_column="avg",
            var_column="var",
            metric_column="metric_name",
            is_binary_column="binary",
            interval_size=0.99,
            power=0.8,
            correction_method=SPOT_1,
        )
        treatment_weights = [5000, 2000, 3000]
        ss = ssc.sample_size(
            treatment_weights=treatment_weights,
            mde_column="mde",
            nim_column="nim",
            preferred_direction_column="preference",
        )

        assert len(ss) == len(df)
        assert ss[REQUIRED_SAMPLE_SIZE_METRIC].values[0] is None
        assert 0.999 < ss[REQUIRED_SAMPLE_SIZE_METRIC].values[1] / 95459 < 1.001
        assert ss[CI_WIDTH].isna().all()



# Copyright 2017-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, Iterable, Tuple

import numpy as np
from bokeh.models import tools
from chartify import Chart
from pandas import DataFrame, concat

from ..abstract_base_classes.confidence_grapher_abc import ConfidenceGrapherABC
from ..confidence_utils import (
    axis_format_precision,
    add_color_column,
    get_remaning_groups,
    get_all_group_columns,
    listify,
    level2str,
    to_finite,
    de_list_if_length_one,
)
from ..constants import (
    POINT_ESTIMATE,
    ORIGINAL_POINT_ESTIMATE,
    DIFFERENCE,
    CI_LOWER,
    CI_UPPER,
    P_VALUE,
    ADJUSTED_LOWER,
    ADJUSTED_UPPER,
    ADJUSTED_P,
    NULL_HYPOTHESIS,
    NIM,
    NIM_TYPE,
    PREFERENCE,
    SFX1,
)
from ...chartgrid import ChartGrid


class ChartifyGrapher(ConfidenceGrapherABC):
    def __init__(
        self,
        data_frame: DataFrame,
        numerator_column: str,
        denominator_column: str,
        categorical_group_columns: str,
        ordinal_group_column: str,
    ):
        self._df = data_frame
        self._numerator = numerator_column
        self._denominator = denominator_column
        self._categorical_group_columns = categorical_group_columns
        self._ordinal_group_column = ordinal_group_column
        self._all_group_columns = get_all_group_columns(self._categorical_group_columns, self._ordinal_group_column)

    def plot_summary(self, summary_df: DataFrame, groupby: Union[str, Iterable]) -> ChartGrid:
        ch = ChartGrid()
        if groupby is None:
            ch.charts.append(self._summary_plot(level_name=None, level_df=summary_df, groupby=groupby))
        else:
            for level_name, level_df in summary_df.groupby(de_list_if_length_one(groupby)):
                ch.charts.append(self._summary_plot(level_name=level_name, level_df=level_df, groupby=groupby))
        return ch

    def plot_difference(
        self,
        difference_df,
        absolute,
        groupby,
        nims: NIM_TYPE,
        use_adjusted_intervals: bool,
        split_plot_by_groups: bool,
    ) -> ChartGrid:
        ch = ChartGrid()
        categorical_groups = get_remaning_groups(listify(groupby), self._ordinal_group_column)

        if len(categorical_groups) == 0 or not split_plot_by_groups:
            ch.charts += self.plot_differece_group(absolute, difference_df, groupby, use_adjusted_intervals).charts
        else:
            for level_name, level_df in difference_df.groupby(categorical_groups):
                ch.charts += self.plot_differece_group(absolute, level_df, groupby, use_adjusted_intervals).charts
        return ch

    def plot_differece_group(self, absolute, difference_df, groupby, use_adjusted_intervals):
        if self._ordinal_group_column in listify(groupby):
            ch = self._ordinal_difference_plot(difference_df, absolute, groupby, use_adjusted_intervals)
            chart_grid = ChartGrid([ch])
        else:
            chart_grid = self._categorical_difference_plot(difference_df, absolute, groupby, use_adjusted_intervals)
        return chart_grid

    def plot_differences(
        self,
        difference_df,
        absolute,
        groupby,
        nims: NIM_TYPE,
        use_adjusted_intervals: bool,
        split_plot_by_groups: bool,
    ) -> ChartGrid:
        ch = ChartGrid()
        categorical_groups = get_remaning_groups(listify(groupby), self._ordinal_group_column)

        if len(categorical_groups) == 0 or not split_plot_by_groups:
            ch.charts += self.plot_differences_group(absolute, difference_df, groupby, use_adjusted_intervals).charts
        else:
            for level_name, level_df in difference_df.groupby(categorical_groups):
                ch.charts += self.plot_differences_group(absolute, level_df, groupby, use_adjusted_intervals).charts
        return ch

    def plot_differences_group(self, absolute, difference_df, groupby, use_adjusted_intervals):
        categorical_groups = get_remaning_groups(groupby, self._ordinal_group_column)
        groupby_columns = self._add_level_columns(categorical_groups)
        if self._ordinal_group_column in listify(groupby):
            ch = self._ordinal_difference_plot(difference_df, absolute, groupby_columns, use_adjusted_intervals)
            chart_grid = ChartGrid([ch])
        else:
            chart_grid = self._categorical_difference_plot(
                difference_df, absolute, groupby_columns, use_adjusted_intervals
            )
        return chart_grid

    def plot_multiple_difference(
        self,
        difference_df,
        absolute,
        groupby,
        level_as_reference,
        nims: NIM_TYPE,
        use_adjusted_intervals: bool,
        split_plot_by_groups: bool,
    ) -> ChartGrid:
        ch = ChartGrid()
        categorical_groups = get_remaning_groups(listify(groupby), self._ordinal_group_column)

        groupby = de_list_if_length_one(groupby)

        if len(categorical_groups) == 0 or not split_plot_by_groups:
            ch.charts += self.plot_multiple_difference_group(
                absolute, difference_df, groupby, level_as_reference, use_adjusted_intervals
            ).charts
        else:
            for level_name, level_df in difference_df.groupby(de_list_if_length_one(categorical_groups)):
                ch.charts += self.plot_multiple_difference_group(
                    absolute, level_df, groupby, level_as_reference, use_adjusted_intervals
                ).charts
        return ch

    def plot_multiple_difference_group(
        self, absolute, difference_df, groupby, level_as_reference, use_adjusted_intervals
    ):
        if self._ordinal_group_column in listify(groupby):
            ch = self._ordinal_multiple_difference_plot(
                difference_df, absolute, groupby, level_as_reference, use_adjusted_intervals
            )
            chart_grid = ChartGrid([ch])
        else:
            chart_grid = self._categorical_multiple_difference_plot(
                difference_df, absolute, groupby, level_as_reference, use_adjusted_intervals
            )
        return chart_grid

    def _ordinal_difference_plot(
        self, difference_df: DataFrame, absolute: bool, groupby: Union[str, Iterable], use_adjusted_intervals: bool
    ) -> Chart:
        remaining_groups = get_remaning_groups(groupby, self._ordinal_group_column)

        if "level_1" in groupby and "level_2" in groupby:
            title = "Change from level_1 to level_2"
        else:
            title = "Change from {} to {}".format(
                difference_df["level_1"].values[0], difference_df["level_2"].values[0]
            )

        y_axis_label = self._get_difference_plot_label(absolute)
        ch = self._ordinal_plot(
            DIFFERENCE,
            difference_df,
            groupby=None,
            level_name="",
            remaining_groups=remaining_groups,
            absolute=absolute,
            title=title,
            y_axis_label=y_axis_label,
            use_adjusted_intervals=use_adjusted_intervals,
        )
        ch.callout.line(0)

        return ch

    def _get_difference_plot_label(self, absolute):
        change_type = "Absolute" if absolute else "Relative"
        return change_type + " change in {} / {}".format(self._numerator, self._denominator)

    def _categorical_difference_plot(
        self, difference_df: DataFrame, absolute: bool, groupby: Union[str, Iterable], use_adjusted_intervals: bool
    ) -> ChartGrid:
        if groupby is None:
            groupby = "dummy_groupby"
            difference_df.loc[:, groupby] = "Difference"

        if "level_1" in groupby and "level_2" in groupby:
            title = "Change from level_1 to level_2"
        else:
            title = "Change from {} to {}".format(
                difference_df["level_1"].values[0], difference_df["level_2"].values[0]
            )
        x_label = "" if groupby is None else "{}".format(groupby)

        chart_grid = self._categorical_difference_chart(
            absolute, difference_df, groupby, title, x_label, use_adjusted_intervals
        )

        return chart_grid

    def _categorical_difference_chart(
        self,
        absolute: bool,
        difference_df: DataFrame,
        groupby_columns: Union[str, Iterable],
        title: str,
        x_label: str,
        use_adjusted_intervals: bool,
    ) -> ChartGrid:
        LOWER, UPPER = (ADJUSTED_LOWER, ADJUSTED_UPPER) if use_adjusted_intervals else (CI_LOWER, CI_UPPER)
        axis_format, y_min, y_max = axis_format_precision(
            numbers=concat(
                [
                    difference_df[LOWER],
                    difference_df[DIFFERENCE],
                    difference_df[UPPER],
                    difference_df[NULL_HYPOTHESIS] if NULL_HYPOTHESIS in difference_df.columns else None,
                ],
            ),
            absolute=absolute,
        )

        df = (
            difference_df.assign(**{LOWER: to_finite(difference_df[LOWER], y_min, y_max)})
            .assign(**{UPPER: to_finite(difference_df[UPPER], y_min, y_max)})
            .assign(level_1=difference_df.level_1.map(level2str))
            .assign(level_2=difference_df.level_2.map(level2str))
            .set_index(groupby_columns)
            .assign(categorical_x=lambda df: df.index.to_numpy())
            .reset_index()
        )

        ch = Chart(x_axis_type="categorical")
        ch.plot.interval(
            data_frame=df.sort_values(groupby_columns),
            categorical_columns=groupby_columns,
            lower_bound_column=LOWER,
            upper_bound_column=UPPER,
            middle_column=DIFFERENCE,
            categorical_order_by="labels",
            categorical_order_ascending=False,
        )
        # Also plot transparent circles, just to be able to show hover box
        ch.style.color_palette.reset_palette_order()
        ch.figure.circle(
            source=df, x="categorical_x", y=DIFFERENCE, size=20, name="center", line_alpha=0, fill_alpha=0
        )
        if NULL_HYPOTHESIS in df.columns and df[NIM].notna().any():
            ch.style.color_palette.reset_palette_order()
            dash_source = (
                df[~df[NIM].isna()]
                .assign(
                    color_column=lambda df: df.apply(
                        lambda row: "red"
                        if row[LOWER] < row[NULL_HYPOTHESIS]
                        and row[PREFERENCE] == "increase"
                        or row[NULL_HYPOTHESIS] < row[UPPER]
                        and row[PREFERENCE] == "decrease"
                        else "green",
                        axis=1,
                    )
                )
                .sort_values(groupby_columns)
            )
            ch.figure.dash(
                source=dash_source,
                x="categorical_x",
                y=NULL_HYPOTHESIS,
                size=320 / len(df),
                line_width=3,
                name="nim",
                line_color="color_column",
            )
        ch.axes.set_yaxis_label(self._get_difference_plot_label(absolute))
        ch.set_source_label("")
        ch.callout.line(0)
        ch.axes.set_yaxis_range(y_min - 0.05 * (y_max - y_min), y_max + 0.05 * (y_max - y_min))
        ch.axes.set_yaxis_tick_format(axis_format)
        ch.set_title(title)
        ch.axes.set_xaxis_label(x_label)
        ch.set_subtitle("")

        self.add_tools(
            chart=ch,
            df=(
                difference_df.set_index(groupby_columns)
                .assign(categorical_x=lambda df: df.index.to_numpy())
                .reset_index()
            ),
            center_name=DIFFERENCE,
            absolute=absolute,
            ordinal=False,
            use_adjusted_intervals=use_adjusted_intervals,
        )

        chart_grid = ChartGrid()
        chart_grid.charts.append(ch)

        return chart_grid

    def _summary_plot(self, level_name: Union[str, Tuple], level_df: DataFrame, groupby: Union[str, Iterable]):
        remaining_groups = get_remaning_groups(self._all_group_columns, groupby)
        if self._ordinal_group_column is not None and self._ordinal_group_column in remaining_groups:
            ch = self._ordinal_summary_plot(level_name, level_df, remaining_groups, groupby)
        else:
            ch = self._categorical_summary_plot(level_name, level_df, remaining_groups, groupby)
        return ch

    def _ordinal_summary_plot(
        self,
        level_name: Union[str, Tuple],
        level_df: DataFrame,
        remaining_groups: Union[str, Iterable],
        groupby: Union[str, Iterable],
    ):
        remaining_groups = get_remaning_groups(remaining_groups, self._ordinal_group_column)
        title = "Estimate of {} / {}".format(self._numerator, self._denominator)
        y_axis_label = "{} / {}".format(self._numerator, self._denominator)
        return self._ordinal_plot(
            POINT_ESTIMATE,
            level_df,
            groupby,
            level_name,
            remaining_groups,
            absolute=True,
            title=title,
            y_axis_label=y_axis_label,
            use_adjusted_intervals=False,
        )

    def _ordinal_plot(
        self,
        center_name: str,
        level_df: DataFrame,
        groupby: Union[str, Iterable],
        level_name: Union[str, Tuple],
        remaining_groups: Union[str, Iterable],
        absolute: bool,
        title: str,
        y_axis_label: str,
        use_adjusted_intervals: bool,
    ):
        LOWER, UPPER = (ADJUSTED_LOWER, ADJUSTED_UPPER) if use_adjusted_intervals else (CI_LOWER, CI_UPPER)
        df = add_color_column(level_df, remaining_groups)
        colors = "color" if remaining_groups else None
        axis_format, y_min, y_max = axis_format_precision(
            numbers=concat(
                [df[LOWER], df[center_name], df[UPPER], df[NULL_HYPOTHESIS] if NULL_HYPOTHESIS in df.columns else None]
            ),
            absolute=absolute,
        )
        ch = Chart(x_axis_type=self._ordinal_type())
        ch.plot.line(
            data_frame=df.sort_values(self._ordinal_group_column),
            x_column=self._ordinal_group_column,
            y_column=center_name,
            color_column=colors,
        )
        # Also plot transparent circles, just to be able to show hover box
        ch.figure.line(source=df, x=self._ordinal_group_column, y=center_name, name="center", line_alpha=0)
        ch.style.color_palette.reset_palette_order()
        ch.plot.area(
            data_frame=(
                df.assign(**{LOWER: to_finite(df[LOWER], y_min, y_max)})
                .assign(**{UPPER: to_finite(df[UPPER], y_min, y_max)})
                .sort_values(self._ordinal_group_column)
            ),
            x_column=self._ordinal_group_column,
            y_column=LOWER,
            second_y_column=UPPER,
            color_column=colors,
        )
        if NULL_HYPOTHESIS in df.columns:
            ch.style.color_palette.reset_palette_order()
            ch.plot.line(
                data_frame=df.sort_values(self._ordinal_group_column),
                x_column=self._ordinal_group_column,
                y_column=NULL_HYPOTHESIS,
                color_column=colors,
                line_dash="dashed",
                line_width=1,
            )
            # Also plot named transparent line, just to be able to show hover box
            ch.figure.line(
                source=df.sort_values(self._ordinal_group_column),
                x=self._ordinal_group_column,
                y=NULL_HYPOTHESIS,
                line_width=3,
                line_alpha=0,
                name="nim",
            )
        ch.axes.set_yaxis_label(y_axis_label)
        ch.axes.set_xaxis_label(self._ordinal_group_column)
        ch.set_source_label("")
        ch.axes.set_yaxis_range(y_min - 0.05 * (y_max - y_min), y_max + 0.05 * (y_max - y_min))
        ch.axes.set_yaxis_tick_format(axis_format)
        subtitle = "" if not groupby else "{}: {}".format(groupby, level_name)
        ch.set_subtitle(subtitle)
        ch.set_title(title)
        if colors:
            ch.set_legend_location("outside_bottom")
        self.add_tools(
            chart=ch,
            df=df,
            center_name=center_name,
            absolute=absolute,
            ordinal=True,
            use_adjusted_intervals=use_adjusted_intervals,
        )
        return ch

    def _categorical_summary_plot(self, level_name, summary_df, remaining_groups, groupby):
        if not remaining_groups:
            remaining_groups = listify(groupby)
        df = summary_df.set_index(remaining_groups).assign(categorical_x=lambda df: df.index.to_numpy()).reset_index()

        axis_format, y_min, y_max = axis_format_precision(
            numbers=concat([df[CI_LOWER], df[POINT_ESTIMATE], df[CI_UPPER]]), absolute=True
        )

        ch = Chart(x_axis_type="categorical")
        ch.plot.interval(
            (
                df.assign(**{CI_LOWER: to_finite(df[CI_LOWER], y_min, y_max)}).assign(
                    **{CI_UPPER: to_finite(df[CI_UPPER], y_min, y_max)}
                )
            ),
            categorical_columns=remaining_groups,
            lower_bound_column=CI_LOWER,
            upper_bound_column=CI_UPPER,
            middle_column=POINT_ESTIMATE,
            categorical_order_by="labels",
            categorical_order_ascending=True,
        )
        # Also plot transparent circles, just to be able to show hover box
        ch.style.color_palette.reset_palette_order()
        ch.figure.circle(
            source=df, x="categorical_x", y=POINT_ESTIMATE, size=20, name="center", line_alpha=0, fill_alpha=0
        )
        ch.set_title("Estimate of {} / {}".format(self._numerator, self._denominator))
        if groupby:
            ch.set_subtitle("{}: {}".format(groupby, level_name))
        else:
            ch.set_subtitle("")
        ch.axes.set_xaxis_label("{}".format(", ".join(remaining_groups)))
        ch.axes.set_yaxis_label("{} / {}".format(self._numerator, self._denominator))
        ch.set_source_label("")
        ch.axes.set_yaxis_tick_format(axis_format)
        self.add_tools(
            chart=ch, df=df, center_name=POINT_ESTIMATE, absolute=True, ordinal=False, use_adjusted_intervals=False
        )
        return ch

    def _ordinal_type(self):
        ordinal_column_type = self._df[self._ordinal_group_column].dtype.type
        axis_type = "datetime" if issubclass(ordinal_column_type, np.datetime64) else "linear"
        return axis_type

    def _ordinal_multiple_difference_plot(
        self,
        difference_df: DataFrame,
        absolute: bool,
        groupby: Union[str, Iterable],
        level_as_reference: bool,
        use_adjusted_intervals: bool,
    ):
        remaining_groups = get_remaning_groups(groupby, self._ordinal_group_column)
        groupby_columns = self._add_level_column(remaining_groups, level_as_reference)
        title = self._get_multiple_difference_title(difference_df, level_as_reference)
        y_axis_label = self._get_difference_plot_label(absolute)
        ch = self._ordinal_plot(
            DIFFERENCE,
            difference_df,
            groupby=None,
            level_name="",
            remaining_groups=groupby_columns,
            absolute=absolute,
            title=title,
            y_axis_label=y_axis_label,
            use_adjusted_intervals=use_adjusted_intervals,
        )
        ch.callout.line(0)
        return ch

    def _categorical_multiple_difference_plot(
        self,
        difference_df: DataFrame,
        absolute: bool,
        groupby: Union[str, Iterable],
        level_as_reference: bool,
        use_adjusted_intervals: bool,
    ):
        groupby_columns = self._add_level_column(groupby, level_as_reference)
        title = self._get_multiple_difference_title(difference_df, level_as_reference)
        x_label = "" if groupby is None else "{}".format(groupby)
        chart_grid = self._categorical_difference_chart(
            absolute, difference_df, groupby_columns, title, x_label, use_adjusted_intervals
        )

        return chart_grid

    def _get_multiple_difference_title(self, difference_df, level_as_reference):
        reference_level = "level_1" if level_as_reference else "level_2"
        title = "Comparison to {}".format(difference_df[reference_level].values[0])
        return title

    def _add_level_column(self, groupby, level_as_reference):
        level_column = "level_2" if level_as_reference else "level_1"
        if groupby is None:
            groupby_columns = level_column
        else:
            if isinstance(groupby, str):
                groupby_columns = [groupby, level_column]
            else:
                groupby_columns = groupby + [level_column]
        return groupby_columns

    def _add_level_columns(self, groupby):
        levels = ["level_1", "level_2"]
        if groupby is None:
            groupby_columns = levels
        else:
            if isinstance(groupby, str):
                groupby_columns = [groupby] + levels
            else:
                groupby_columns = groupby + levels
        return groupby_columns

    def add_ci_to_chart_datasources(
        self, chart: Chart, df: DataFrame, center_name: str, ordinal: bool, use_adjusted_intervals: bool
    ):
        LOWER, UPPER = (ADJUSTED_LOWER, ADJUSTED_UPPER) if use_adjusted_intervals else (CI_LOWER, CI_UPPER)
        group_col = "color" if ordinal and "color" in df.columns else "categorical_x"
        for data in chart.data:
            if center_name in data.keys() or NULL_HYPOTHESIS in data.keys():
                index = data["index"]
                data[LOWER] = np.array(df[LOWER][index])
                data[UPPER] = np.array(df[UPPER][index])
                data["color"] = np.array(df[group_col][index])
            if DIFFERENCE in data.keys() or NULL_HYPOTHESIS in data.keys():
                index = data["index"]
                data["reference_level"] = np.array(df["level_1"][index])
                data[DIFFERENCE] = np.array(df[DIFFERENCE][index])
                data["p_value"] = np.array(df[P_VALUE][index])
                data["adjusted_p"] = np.array(df[ADJUSTED_P][index])
                data["reference_level_avg"] = np.array(df[ORIGINAL_POINT_ESTIMATE + SFX1][index])
                if NULL_HYPOTHESIS in df.columns:
                    data["null_hyp"] = np.array(df[NULL_HYPOTHESIS][index])

    def add_tools(
        self,
        chart: Chart,
        df: DataFrame,
        center_name: str,
        absolute: bool,
        ordinal: bool,
        use_adjusted_intervals: bool,
    ):
        self.add_ci_to_chart_datasources(chart, df, center_name, ordinal, use_adjusted_intervals)
        LOWER, UPPER = (ADJUSTED_LOWER, ADJUSTED_UPPER) if use_adjusted_intervals else (CI_LOWER, CI_UPPER)

        if len(chart.figure.legend) > 0:
            chart.figure.legend.click_policy = "hide"
        axis_format, y_min, y_max = axis_format_precision(
            numbers=concat(
                [df[LOWER], df[center_name], df[UPPER], df[NULL_HYPOTHESIS] if NULL_HYPOTHESIS in df.columns else None]
            ),
            absolute=absolute,
            extra_zeros=2,
        )
        axis_format_reference_level, _, _ = axis_format_precision(
            numbers=concat(
                [df[LOWER], df[center_name], df[UPPER], df[NULL_HYPOTHESIS] if NULL_HYPOTHESIS in df.columns else None]
            ),
            absolute=True,
            extra_zeros=2,
        )
        ordinal_tool_tip = [] if not ordinal else [(self._ordinal_group_column, f"@{self._ordinal_group_column}")]
        p_value_tool_tip = (
            (
                [("p-value", "@p_value{0.0000}")]
                + ([("adjusted p-value", "@adjusted_p{0.0000}")] if len(df) > 1 else [])
            )
            if center_name == DIFFERENCE
            else []
        )
        nim_tool_tip = [("null hypothesis", f"@null_hyp{{{axis_format}}}")] if NULL_HYPOTHESIS in df.columns else []
        reference_level_tool_tip = (
            [("reference level", f"@reference_level: @reference_level_avg{{{axis_format_reference_level}}}")]
            if "level_1" in df.columns
            else []
        )
        tooltips = (
            [("group", "@color")]
            + reference_level_tool_tip
            + ordinal_tool_tip
            + [(f"{center_name}", f"@{center_name}{{{axis_format}}}")]
            + [
                (
                    ("adjusted " if use_adjusted_intervals else "") + "confidence interval",
                    f"(@{{{LOWER}}}{{{axis_format}}}," f" @{{{UPPER}}}{{{axis_format}}})",
                )
            ]
            + p_value_tool_tip
            + nim_tool_tip
        )
        lines_with_hover = ["center", "nim"]
        renderers = [r for r in chart.figure.renderers if r.name in lines_with_hover]
        hover = tools.HoverTool(tooltips=tooltips, renderers=renderers)

        box_zoom = tools.BoxZoomTool()

        chart.figure.add_tools(
            hover, tools.ZoomInTool(), tools.ZoomOutTool(), box_zoom, tools.PanTool(), tools.ResetTool()
        )
        chart.figure.toolbar.active_drag = box_zoom



from typing import Tuple, Dict

import numpy as np
from pandas import DataFrame, Series
from statsmodels.stats.proportion import proportion_confint, proportions_chisquare, confint_proportions_2indep

from spotify_confidence.analysis.confidence_utils import power_calculation
from spotify_confidence.analysis.constants import (
    NUMERATOR,
    DENOMINATOR,
    INTERVAL_SIZE,
    POINT_ESTIMATE,
    VARIANCE,
    CI_LOWER,
    CI_UPPER,
    SFX1,
    SFX2,
)


def point_estimate(df: DataFrame, **kwargs: Dict[str, str]) -> float:
    numerator = kwargs[NUMERATOR]
    denominator = kwargs[DENOMINATOR]
    if (df[denominator] == 0).any():
        raise ValueError("""Can't compute point estimate: denominator is 0""")
    return df[numerator] / df[denominator]


def variance(df: DataFrame, **kwargs: Dict[str, str]) -> Series:
    variance = df[POINT_ESTIMATE] * (1 - df[POINT_ESTIMATE])
    if (variance < 0).any():
        raise ValueError(f"Computed variance is negative: {variance}. " "Please check your inputs.")
    return variance


def std_err(df: DataFrame, **kwargs: Dict[str, str]) -> Series:
    denominator = kwargs[DENOMINATOR]
    return np.sqrt(df[VARIANCE + SFX1] / df[denominator + SFX1] + df[VARIANCE + SFX2] / df[denominator + SFX2])


def add_point_estimate_ci(df: DataFrame, **kwargs: Dict[str, str]) -> Series:
    numerator = kwargs[NUMERATOR]
    denominator = kwargs[DENOMINATOR]
    interval_size = kwargs[INTERVAL_SIZE]
    df[CI_LOWER], df[CI_UPPER] = proportion_confint(
        count=df[numerator],
        nobs=df[denominator],
        alpha=1 - interval_size,
    )
    return df


def p_value(df: DataFrame, **kwargs: Dict[str, str]) -> Series:
    n1, n2 = kwargs[NUMERATOR] + SFX1, kwargs[NUMERATOR] + SFX2
    d1, d2 = kwargs[DENOMINATOR] + SFX1, kwargs[DENOMINATOR] + SFX2

    def p_value_row(row):
        _, p_value, _ = proportions_chisquare(
            count=[row[n1], row[n2]],
            nobs=[row[d1], row[d2]],
        )
        return p_value

    return df.apply(p_value_row, axis=1)


def ci(df: DataFrame, alpha_column: str, **kwargs: Dict[str, str]) -> Tuple[Series, Series]:
    n1, n2 = kwargs[NUMERATOR] + SFX1, kwargs[NUMERATOR] + SFX2
    d1, d2 = kwargs[DENOMINATOR] + SFX1, kwargs[DENOMINATOR] + SFX2
    return confint_proportions_2indep(
        count1=df[n2],
        nobs1=df[d2],
        count2=df[n1],
        nobs2=df[d1],
        alpha=df[alpha_column],
        compare="diff",
        method="wald",
    )


def achieved_power(df: DataFrame, mde: float, alpha: float, **kwargs: Dict[str, str]) -> DataFrame:
    n1, n2 = kwargs[NUMERATOR] + SFX1, kwargs[NUMERATOR] + SFX2
    d1, d2 = kwargs[DENOMINATOR] + SFX1, kwargs[DENOMINATOR] + SFX2

    pooled_prop = (df[n1] + df[n2]) / (df[d1] + df[d2])
    var_pooled = pooled_prop * (1 - pooled_prop)

    return power_calculation(mde, var_pooled, alpha, df[d1], df[d2])



# Copyright 2017-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, Iterable, List, Tuple, Dict

import numpy as np
from pandas import DataFrame, Series
from scipy import stats as st

from spotify_confidence.analysis.confidence_utils import (
    get_all_categorical_group_columns,
    get_all_group_columns,
    validate_data,
    remove_group_columns,
    groupbyApplyParallel,
    is_non_inferiority,
    reset_named_indices,
    de_list_if_length_one,
)
from spotify_confidence.analysis.constants import (
    INTERVAL_SIZE,
    POINT_ESTIMATE,
    FINAL_EXPECTED_SAMPLE_SIZE,
    MDE,
    CORRECTION_METHOD,
    VARIANCE,
    NUMBER_OF_COMPARISONS,
    TREATMENT_WEIGHTS,
    IS_BINARY,
    CI_LOWER,
    CI_UPPER,
    DIFFERENCE,
    SFX1,
    ADJUSTED_ALPHA_POWER_SAMPLE_SIZE,
    POWER,
    POWERED_EFFECT,
    ADJUSTED_POWER,
    ADJUSTED_LOWER,
    ADJUSTED_UPPER,
    REQUIRED_SAMPLE_SIZE_METRIC,
    OPTIMAL_KAPPA,
    OPTIMAL_WEIGHTS,
    CI_WIDTH,
    NULL_HYPOTHESIS,
    ALTERNATIVE_HYPOTHESIS,
    NIM,
    PREFERENCE_TEST,
    TWO_SIDED,
    CORRECTION_METHODS,
    ZTEST,
    ORIGINAL_POINT_ESTIMATE,
    ORIGINAL_VARIANCE,
)
from spotify_confidence.analysis.frequentist.confidence_computers import confidence_computers
from spotify_confidence.analysis.frequentist.multiple_comparison import (
    get_num_comparisons,
    set_alpha_and_adjust_preference,
    get_preference,
    add_adjusted_power,
)
from spotify_confidence.analysis.frequentist.nims_and_mdes import (
    add_nims_and_mdes,
)


class SampleSizeComputer:
    def __init__(
        self,
        data_frame: DataFrame,
        categorical_group_columns: Union[str, Iterable],
        interval_size: float,
        correction_method: str,
        metric_column: str,
        power: float,
        point_estimate_column: str,
        var_column: str,
        is_binary_column: str,
    ):
        self._df = data_frame.reset_index(drop=True)
        self._point_estimate_column = point_estimate_column
        self._var_column = var_column
        self._is_binary = is_binary_column
        if self._point_estimate_column is not None and self._var_column is not None and self._is_binary is not None:
            mean = self._df.query(f"{self._is_binary} == True")[self._point_estimate_column]
            var = self._df.query(f"{self._is_binary} == True")[self._var_column]
            if not np.allclose(var, (mean * (1 - mean)), equal_nan=True):
                raise ValueError(
                    f"{var_column} doesn't equal {point_estimate_column}*(1-{point_estimate_column}) "
                    f"for all binary rows. Please check your data."
                )

        self._categorical_group_columns = get_all_categorical_group_columns(
            categorical_group_columns, metric_column, None
        )
        self._segments = remove_group_columns(self._categorical_group_columns, metric_column)
        self._segments = remove_group_columns(self._segments, None)
        self._metric_column = metric_column
        self._interval_size = interval_size
        self._power = power

        if correction_method.lower() not in CORRECTION_METHODS:
            raise ValueError(f"Use one of the correction methods in {CORRECTION_METHODS}")
        self._correction_method = correction_method

        self._single_metric = False
        if self._metric_column is not None and data_frame.groupby(self._metric_column, sort=False).ngroups == 1:
            self._single_metric = True

        self._all_group_columns = get_all_group_columns(self._categorical_group_columns, None)

        columns_that_must_exist = [self._point_estimate_column, self._var_column]
        validate_data(self._df, columns_that_must_exist, self._all_group_columns, None)

        self._sufficient = None

    def compute_summary(self, verbose: bool) -> DataFrame:
        return (
            self._sufficient_statistics
            if verbose
            else self._sufficient_statistics[
                self._all_group_columns
                + ([self._metric_column] if self._metric_column is not None and self._single_metric else [])
                + [c for c in [self._numerator, self._denominator] if c is not None]
                + [POINT_ESTIMATE, CI_LOWER, CI_UPPER]
            ]
        )

    @property
    def _sufficient_statistics(self) -> DataFrame:
        if self._sufficient is None:
            groupby = [self._metric_column]
            self._sufficient = (
                self._df.groupby(groupby, sort=False, group_keys=True)
                .apply(
                    lambda df: df.assign(**{POINT_ESTIMATE: lambda df: df[self._point_estimate_column]})
                    .assign(**{ORIGINAL_POINT_ESTIMATE: lambda df: df[self._point_estimate_column]})
                    .assign(**{VARIANCE: lambda df: df[self._var_column]})
                    .assign(**{ORIGINAL_VARIANCE: lambda df: df[self._var_column]})
                )
                .pipe(reset_named_indices)
            )
        return self._sufficient

    def compute_sample_size(
        self,
        treatment_weights: Iterable,
        mde_column: str,
        nim_column: str,
        preferred_direction_column: str,
        final_expected_sample_size_column: str,
    ) -> DataFrame:
        kwargs, group_columns, sample_size_df = self._initialise_sample_size_and_power_computation(
            final_expected_sample_size_column, mde_column, nim_column, preferred_direction_column, treatment_weights
        )
        sample_size_df = groupbyApplyParallel(
            sample_size_df.pipe(set_alpha_and_adjust_preference, **kwargs).groupby(
                de_list_if_length_one(group_columns),
                as_index=False,
                sort=False,
            ),
            lambda df: _compute_sample_sizes_and_ci_widths(df, **kwargs),
        )

        return sample_size_df.reset_index()

    def compute_powered_effect(
        self,
        treatment_weights: Iterable,
        mde_column: str,
        nim_column: str,
        preferred_direction_column: str,
        sample_size: float,
    ) -> DataFrame:
        kwargs, group_columns, powered_effect_df = self._initialise_sample_size_and_power_computation(
            sample_size, mde_column, nim_column, preferred_direction_column, treatment_weights
        )
        powered_effect_df = groupbyApplyParallel(
            powered_effect_df.pipe(set_alpha_and_adjust_preference, **kwargs).groupby(
                de_list_if_length_one(group_columns),
                as_index=False,
                sort=False,
            ),
            lambda df: _compute_powered_effects(df, **kwargs),
        )

        return powered_effect_df.reset_index()

    def _initialise_sample_size_and_power_computation(
        self, final_expected_sample_size_column, mde_column, nim_column, preferred_direction_column, treatment_weights
    ):
        sample_size_df = (
            self._sufficient_statistics.pipe(
                lambda df: df if self._all_group_columns == [] else df.set_index(self._all_group_columns)
            )
            .pipe(
                add_nims_and_mdes,
                mde_column=mde_column,
                nim_column=nim_column,
                preferred_direction_column=preferred_direction_column,
            )
            .assign(**{PREFERENCE_TEST: lambda df: get_preference(df, self._correction_method)})
            .assign(**{POWER: self._power})
            .pipe(
                add_adjusted_power,
                correction_method=self._correction_method,
                metric_column=self._metric_column,
                single_metric=self._single_metric,
            )
        )
        group_columns = [column for column in sample_size_df.index.names if column is not None]
        n_comparisons = get_num_comparisons(
            sample_size_df,
            self._correction_method,
            number_of_level_comparisons=len(treatment_weights) - 1,
            groupby=group_columns,
            metric_column=self._metric_column,
            treatment_column=None,
            single_metric=self._single_metric,
            segments=self._segments,
        )
        kwargs = {
            MDE: mde_column,
            NUMBER_OF_COMPARISONS: n_comparisons,
            TREATMENT_WEIGHTS: treatment_weights,
            INTERVAL_SIZE: self._interval_size,
            CORRECTION_METHOD: self._correction_method,
            IS_BINARY: self._is_binary,
            FINAL_EXPECTED_SAMPLE_SIZE: final_expected_sample_size_column,
        }
        return kwargs, group_columns, sample_size_df

    def compute_optimal_weights_and_sample_size(
        self, sample_size_df: DataFrame, number_of_groups: int
    ) -> Tuple[Iterable, int]:
        sample_size_df = (
            sample_size_df.reset_index(drop=True)
            .assign(**{OPTIMAL_KAPPA: lambda df: df.apply(_optimal_kappa, is_binary_column=self._is_binary, axis=1)})
            .assign(
                **{
                    OPTIMAL_WEIGHTS: lambda df: df.apply(
                        lambda row: _optimal_weights(row[OPTIMAL_KAPPA], number_of_groups), axis=1
                    )
                }
            )
        )

        group_columns = [column for column in sample_size_df.index.names if column is not None] + [self._metric_column]
        kwargs = {
            IS_BINARY: self._is_binary,
        }
        return _find_optimal_group_weights_across_rows(sample_size_df, number_of_groups, group_columns, **kwargs)


def _adjust_if_absolute(df: DataFrame, absolute: bool) -> DataFrame:
    if absolute:
        return df.assign(absolute_difference=absolute)
    else:
        return (
            df.assign(absolute_difference=absolute)
            .assign(**{DIFFERENCE: df[DIFFERENCE] / df[POINT_ESTIMATE + SFX1]})
            .assign(**{CI_LOWER: df[CI_LOWER] / df[POINT_ESTIMATE + SFX1]})
            .assign(**{CI_UPPER: df[CI_UPPER] / df[POINT_ESTIMATE + SFX1]})
            .assign(**{ADJUSTED_LOWER: df[ADJUSTED_LOWER] / df[POINT_ESTIMATE + SFX1]})
            .assign(**{ADJUSTED_UPPER: df[ADJUSTED_UPPER] / df[POINT_ESTIMATE + SFX1]})
            .assign(**{NULL_HYPOTHESIS: df[NULL_HYPOTHESIS] / df[POINT_ESTIMATE + SFX1]})
            .assign(**{POWERED_EFFECT: df[POWERED_EFFECT] / df[POINT_ESTIMATE + SFX1]})
        )


def _compute_sample_sizes_and_ci_widths(df: DataFrame, **kwargs: Dict) -> DataFrame:
    return df.pipe(_sample_size_from_summary_df, **kwargs).pipe(_ci_width, **kwargs)


def _sample_size_from_summary_df(df: DataFrame, **kwargs: Dict) -> DataFrame:
    if (df[ADJUSTED_POWER].isna()).any():
        df[REQUIRED_SAMPLE_SIZE_METRIC] = None
    else:
        all_weights = kwargs[TREATMENT_WEIGHTS]
        control_weight, treatment_weights = all_weights[0], all_weights[1:]

        binary = df[kwargs[IS_BINARY]].values[0]
        z_alpha = st.norm.ppf(
            1
            - df[ADJUSTED_ALPHA_POWER_SAMPLE_SIZE].values[0] / (2 if df[PREFERENCE_TEST].values[0] == TWO_SIDED else 1)
        )
        z_power = st.norm.ppf(df[ADJUSTED_POWER].values[0])
        non_inferiority = is_non_inferiority(df[NIM].values[0])

        max_sample_size = 0
        for treatment_weight in treatment_weights:
            kappa = control_weight / treatment_weight
            proportion_of_total = (control_weight + treatment_weight) / sum(all_weights)

            if ALTERNATIVE_HYPOTHESIS in df and NULL_HYPOTHESIS in df and (df[ALTERNATIVE_HYPOTHESIS].notna()).all():
                this_sample_size = confidence_computers[ZTEST].required_sample_size(
                    proportion_of_total=proportion_of_total,
                    z_alpha=z_alpha,
                    z_power=z_power,
                    binary=binary,
                    non_inferiority=non_inferiority,
                    hypothetical_effect=df[ALTERNATIVE_HYPOTHESIS] - df[NULL_HYPOTHESIS],
                    control_avg=df[POINT_ESTIMATE],
                    control_var=df[VARIANCE],
                    kappa=kappa,
                )
                max_sample_size = max(this_sample_size.max(), max_sample_size)

        df[REQUIRED_SAMPLE_SIZE_METRIC] = None if max_sample_size == 0 else max_sample_size

    return df


def _compute_powered_effects(df: DataFrame, **kwargs: Dict) -> DataFrame:
    return df.pipe(_powered_effect_from_summary_df, **kwargs)


def _powered_effect_from_summary_df(df: DataFrame, **kwargs: Dict) -> DataFrame:
    if (df[ADJUSTED_POWER].isna()).any():
        df[REQUIRED_SAMPLE_SIZE_METRIC] = None
    else:
        all_weights = kwargs[TREATMENT_WEIGHTS]
        control_weight, treatment_weights = all_weights[0], all_weights[1:]

        current_number_of_units = kwargs[FINAL_EXPECTED_SAMPLE_SIZE]

        binary = df[kwargs[IS_BINARY]].values[0]
        z_alpha = st.norm.ppf(
            1
            - df[ADJUSTED_ALPHA_POWER_SAMPLE_SIZE].values[0] / (2 if df[PREFERENCE_TEST].values[0] == TWO_SIDED else 1)
        )
        z_power = st.norm.ppf(df[ADJUSTED_POWER].values[0])
        non_inferiority = is_non_inferiority(df[NIM].values[0])

        max_powered_effect = 0
        for treatment_weight in treatment_weights:
            kappa = control_weight / treatment_weight
            proportion_of_total = (control_weight + treatment_weight) / sum(all_weights)

            this_powered_effect = df[POWERED_EFFECT] = confidence_computers[ZTEST].powered_effect(
                df=df.assign(kappa=kappa)
                .assign(current_number_of_units=current_number_of_units)
                .assign(proportion_of_total=proportion_of_total),
                z_alpha=z_alpha,
                z_power=z_power,
                binary=binary,
                non_inferiority=non_inferiority,
                avg_column=POINT_ESTIMATE,
                var_column=VARIANCE,
            )

            max_powered_effect = max(this_powered_effect.max(), max_powered_effect)

        df[POWERED_EFFECT] = None if max_powered_effect == 0 else max_powered_effect

    return df


def _ci_width(df: DataFrame, **kwargs: Dict) -> DataFrame:
    expected_sample_size = (
        None if kwargs[FINAL_EXPECTED_SAMPLE_SIZE] is None else df[kwargs[FINAL_EXPECTED_SAMPLE_SIZE]].values[0]
    )
    if expected_sample_size is None or np.isnan(expected_sample_size):
        return df.assign(**{CI_WIDTH: None})

    all_weights = kwargs[TREATMENT_WEIGHTS]
    control_weight, treatment_weights = all_weights[0], all_weights[1:]
    sum_of_weights = sum(all_weights)

    control_count = int((control_weight / sum_of_weights) * expected_sample_size)
    if control_count == 0:
        return df.assign(**{CI_WIDTH: float("inf")})

    else:
        binary = df[kwargs[IS_BINARY]].values[0]
        z_alpha = st.norm.ppf(
            1
            - df[ADJUSTED_ALPHA_POWER_SAMPLE_SIZE].values[0] / (2 if df[PREFERENCE_TEST].values[0] == TWO_SIDED else 1)
        )

        non_inferiority = is_non_inferiority(df[NIM].values[0])
        max_ci_width = 0
        for treatment_weight in treatment_weights:
            treatment_count = int((treatment_weight / sum_of_weights) * expected_sample_size)
            if treatment_count == 0:
                return df.assign(**{CI_WIDTH: float("inf")})
            else:
                comparison_ci_width = confidence_computers[ZTEST].ci_width(
                    z_alpha=z_alpha,
                    binary=binary,
                    non_inferiority=non_inferiority,
                    hypothetical_effect=df[ALTERNATIVE_HYPOTHESIS] - df[NULL_HYPOTHESIS],
                    control_avg=df[POINT_ESTIMATE],
                    control_var=df[VARIANCE],
                    control_count=control_count,
                    treatment_count=treatment_count,
                )

            max_ci_width = max(comparison_ci_width.max(), max_ci_width)

        df[CI_WIDTH] = None if max_ci_width == 0 else max_ci_width

    return df


def _optimal_kappa(row: Series, is_binary_column) -> float:
    def _binary_variance(p: float) -> float:
        return p * (1 - p)

    if row[is_binary_column]:
        if is_non_inferiority(row[NIM]):
            return 1.0
        else:
            if row[POINT_ESTIMATE] == 0.0:
                # variance will be 0 as well in this case. This if-branch is important to avoid divide by zero problems
                return 1.0
            else:
                hypothetical_effect = row[ALTERNATIVE_HYPOTHESIS] - row[NULL_HYPOTHESIS]
                return np.sqrt(
                    _binary_variance(row[POINT_ESTIMATE]) / _binary_variance(row[POINT_ESTIMATE] + hypothetical_effect)
                )
    else:
        return 1.0


def _optimal_weights(kappa: float, number_of_groups) -> Iterable:
    treatment_weight = 1 / (kappa + number_of_groups - 1)
    control_weight = kappa * treatment_weight
    return [control_weight] + [treatment_weight for _ in range(number_of_groups - 1)]


def _find_optimal_group_weights_across_rows(
    df: DataFrame, group_count: int, group_columns: Iterable, **kwargs: Dict
) -> (List[float], int):
    min_kappa = min(df[OPTIMAL_KAPPA])
    max_kappa = max(df[OPTIMAL_KAPPA])

    if min_kappa == max_kappa:
        optimal_weights = df[OPTIMAL_WEIGHTS][0]
        optimal_sample_size = _calculate_optimal_sample_size_given_weights(
            df, optimal_weights, group_columns, **kwargs
        )
        return optimal_weights, optimal_sample_size

    in_between_kappas = np.linspace(min_kappa, max_kappa, 100)
    min_optimal_sample_size = float("inf")
    optimal_weights = []
    for kappa in in_between_kappas:
        weights = _optimal_weights(kappa, group_count)
        optimal_sample_size = _calculate_optimal_sample_size_given_weights(df, weights, group_columns, **kwargs)
        if optimal_sample_size is not None and optimal_sample_size < min_optimal_sample_size:
            min_optimal_sample_size = optimal_sample_size
            optimal_weights = weights
    min_optimal_sample_size = np.nan if min_optimal_sample_size == 0 else min_optimal_sample_size
    return optimal_weights, min_optimal_sample_size


def _calculate_optimal_sample_size_given_weights(
    df: DataFrame, optimal_weights: List[float], group_columns: Iterable, **kwargs: Dict
) -> int:
    kwargs[TREATMENT_WEIGHTS] = optimal_weights
    sample_size_df = groupbyApplyParallel(
        df.groupby(de_list_if_length_one(group_columns), as_index=False, sort=False),
        lambda df: _sample_size_from_summary_df(df, **kwargs),
    )

    if sample_size_df[REQUIRED_SAMPLE_SIZE_METRIC].isna().all():
        return None
    optimal_sample_size = sample_size_df[REQUIRED_SAMPLE_SIZE_METRIC].max()

    return np.ceil(optimal_sample_size) if np.isfinite(optimal_sample_size) else optimal_sample_size



import numpy as np
import pandas as pd

import spotify_confidence
from spotify_confidence.analysis.constants import REGRESSION_PARAM, DECREASE_PREFFERED, METHOD_COLUMN_NAME


class TestUnivariateSingleMetric(object):
    def setup(self):
        np.random.seed(123)
        n = 10000
        d = np.random.randint(2, size=n)
        x = np.random.standard_normal(size=n) + 100
        y = 0.5 * d + 0.5 * x + np.random.standard_normal(size=n)
        data = pd.DataFrame({"variation_name": list(map(str, d)), "y": y, "x": x})
        data = (
            data.assign(xy=y * x)
            .assign(x2=x**2)
            .assign(y2=y**2)
            .groupby(["variation_name"])
            .agg({"y": ["sum", "count"], "y2": "sum", "x": "sum", "x2": "sum", "xy": "sum"})
            .reset_index()
        )

        data.columns = data.columns.map("_".join).str.strip("_")
        data = data.assign(**{"metric_name": "metricA"})
        self.n = n
        self.x = x
        self.y = y
        self.d = d
        self.data = data

        self.test = spotify_confidence.ZTestLinreg(
            data_frame=data,
            numerator_column="y_sum",
            numerator_sum_squares_column="y2_sum",
            denominator_column="y_count",
            categorical_group_columns=["variation_name"],
            feature_column="x_sum",
            feature_sum_squares_column="x2_sum",
            feature_cross_sum_column="xy_sum",
            interval_size=0.99,
            correction_method="bonferroni",
            metric_column="metric_name",
        )

    def test_summary(self):
        summary_df = self.test.summary(verbose=True)
        print(summary_df)
        assert len(summary_df) == len(self.data)

    def test_parameters_univariate(self):
        def linreg(X, y):
            return np.matmul(np.linalg.inv(np.matmul(np.transpose(X), X)), np.matmul(np.transpose(X), y))

        summary_df = self.test.summary(verbose=True)
        X = np.ones((self.n, 2))
        X[:, 1] = self.x
        b0 = linreg(X, self.y)
        y_adj = self.y - np.matmul(X, b0)
        X[:, 1] = self.d
        b1 = linreg(X, y_adj)
        assert np.allclose(b0[1], summary_df[REGRESSION_PARAM][0], rtol=0.0001)

        diff = self.test.difference(level_1="0", level_2="1", verbose=True, groupby="metric_name")
        assert np.allclose(b1[1], diff["difference"])

        v0 = np.var(y_adj[self.d == 0])
        v1 = np.var(y_adj[self.d == 1])
        n0 = y_adj[self.d == 0].size
        n1 = y_adj[self.d == 1].size
        assert np.allclose(diff["std_err"], np.sqrt(v0 / n0 + v1 / n1), rtol=1e-3)


class TestUnivariateMultiMetric(object):
    def setup(self):
        np.random.seed(123)
        n = 20000
        d = np.random.randint(2, size=n)
        x = np.random.standard_normal(size=n)
        y = 0.5 * d + 0.5 * x + np.random.standard_normal(size=n)
        m = np.repeat(["metricA", "metricB"], int(n / 2))
        data = pd.DataFrame({"variation_name": list(map(str, d)), "metric_name": m, "y": y, "x": x})
        data = (
            data.assign(xy=y * x)
            .assign(x2=x**2)
            .assign(y2=y**2)
            .groupby(["variation_name", "metric_name"])
            .agg({"y": ["sum", "count"], "y2": "sum", "x": "sum", "x2": "sum", "xy": "sum"})
            .reset_index()
        )

        data.columns = data.columns.map("_".join).str.strip("_")

        self.n = n
        self.x = x
        self.y = y
        self.d = d
        self.m = m
        self.data = data

        self.test = spotify_confidence.ZTestLinreg(
            data_frame=data,
            numerator_column="y_sum",
            numerator_sum_squares_column="y2_sum",
            denominator_column="y_count",
            categorical_group_columns=["variation_name"],
            feature_column="x_sum",
            feature_sum_squares_column="x2_sum",
            feature_cross_sum_column="xy_sum",
            interval_size=0.99,
            correction_method="bonferroni",
            metric_column="metric_name",
        )

    def test_summary(self):
        summary_df = self.test.summary(verbose=True)
        print(summary_df)
        assert len(summary_df) == len(self.data)

    def test_parameters_univariate(self):
        def linreg(X, y):
            return np.matmul(np.linalg.inv(np.matmul(np.transpose(X), X)), np.matmul(np.transpose(X), y))

        summary_df = self.test.summary(verbose=True)
        N0 = self.m[self.m == "metricA"].size
        N1 = self.n - N0
        X = np.ones((N0, 2))
        X[:, 1] = self.x[self.m == "metricA"]
        b0 = linreg(X, self.y[self.m == "metricA"])
        y_adj = self.y[self.m == "metricA"] - np.matmul(X, b0)
        X[:, 1] = self.d[self.m == "metricA"]
        b1 = linreg(X, y_adj)
        assert np.allclose(b0[1], summary_df[REGRESSION_PARAM][0], rtol=0.0001)

        diff = self.test.difference(level_1=("0", "metricA"), level_2=("1", "metricA"), verbose=True)
        assert np.allclose(b1[1], diff["difference"])

        idx = self.d[self.m == "metricA"]
        v0 = np.var(y_adj[idx == 0])
        v1 = np.var(y_adj[idx == 1])
        n0 = y_adj[idx == 0].size
        n1 = y_adj[idx == 1].size
        assert np.allclose(diff["std_err"], np.sqrt(v0 / n0 + v1 / n1), rtol=1e-3)

        X = np.ones((N1, 2))
        X[:, 1] = self.x[self.m == "metricB"]
        b0 = linreg(X, self.y[self.m == "metricB"])
        y_adj = self.y[self.m == "metricB"] - np.matmul(X, b0)
        X[:, 1] = self.d[self.m == "metricB"]
        b1 = linreg(X, y_adj)
        assert np.allclose(b0[1], summary_df[REGRESSION_PARAM][1], rtol=0.0001)

        diff = self.test.difference(level_1=("0", "metricB"), level_2=("1", "metricB"), verbose=True)
        assert np.allclose(b1[1], diff["difference"])

        idx = self.d[self.m == "metricB"]
        v0 = np.var(y_adj[idx == 0])
        v1 = np.var(y_adj[idx == 1])
        n0 = y_adj[idx == 0].size
        n1 = y_adj[idx == 1].size
        assert np.allclose(diff["std_err"], np.sqrt(v0 / n0 + v1 / n1), rtol=1e-3)


class TestUnivariateNoFeatures(object):
    def setup(self):
        self.data = pd.DataFrame(
            [
                {
                    "group": "1",
                    "count": 5000,
                    "sum": 10021.0,
                    "sum_of_squares": 25142.0,
                    "avg": 2.004210,
                    "var": 1.0116668,
                },
                {
                    "group": "2",
                    "count": 5000,
                    "sum": 9892.0,
                    "sum_of_squares": 24510.0,
                    "avg": 1.978424,
                    "var": 0.9881132,
                },
            ]
        )

        self.ztest = spotify_confidence.ZTest(
            self.data,
            numerator_column="sum",
            numerator_sum_squares_column="sum_of_squares",
            denominator_column="count",
            categorical_group_columns="group",
            interval_size=0.99,
        )

        self.ztestlinreg = spotify_confidence.ZTestLinreg(
            data_frame=self.data,
            numerator_column="sum",
            numerator_sum_squares_column="sum_of_squares",
            denominator_column="count",
            categorical_group_columns="group",
            feature_column=None,
            feature_sum_squares_column=None,
            feature_cross_sum_column=None,
            interval_size=0.99,
        )

    def test_summary(self):
        summary_ztest = self.ztest.summary(verbose=True).drop([METHOD_COLUMN_NAME], axis=1)
        summary_ztestlinreg = self.ztestlinreg.summary(verbose=True).drop([METHOD_COLUMN_NAME], axis=1)
        pd.testing.assert_frame_equal(summary_ztest, summary_ztestlinreg)


class TestMultivariateSingleMetric(object):
    def setup(self):
        np.random.seed(123)

        n = 10000
        d = np.random.randint(2, size=n)
        x1 = np.random.standard_normal(size=n) + 50
        x2 = np.random.standard_normal(size=n) + 100
        y = 0.5 * d + 0.5 * x1 + 0.5 * x2 + np.random.standard_normal(size=n)
        df = pd.DataFrame({"variation_name": list(map(str, d)), "y": y, "x1": x1, "x2": x2})

        data = df.assign(y2=y**2).groupby(["variation_name"]).agg({"y": ["sum", "count"], "y2": "sum"}).reset_index()

        data.columns = data.columns.map("_".join).str.strip("_")

        def _to_XX(data, features):
            X = data[features].to_numpy()
            XX = np.matmul(np.transpose(X), X)
            return XX

        def _to_Xy(data, features, metric):
            X = data[features].to_numpy()
            y = data[metric].to_numpy()
            Xy = np.matmul(np.transpose(X), y)
            return Xy

        def _to_Xs(data, features):
            X = data[features].to_numpy()
            Xs = np.sum(X, axis=0)
            return Xs

        XX = (
            df.groupby(["variation_name"])
            .apply(lambda x: pd.Series({"out": _to_XX(x, features=["x1", "x2"])}))
            .to_records()["out"]
        )
        Xy = (
            df.groupby(["variation_name"])
            .apply(lambda x: pd.Series({"out": _to_Xy(x, features=["x1", "x2"], metric="y")}))
            .to_records()["out"]
        )
        Xs = (
            df.groupby(["variation_name"])
            .apply(lambda x: pd.Series({"out": _to_Xs(x, features=["x1", "x2"])}))
            .to_records()["out"]
        )

        data["XX"] = XX
        data["Xy"] = Xy
        data["Xs"] = Xs
        data = data.assign(**{"metric_name": "metricA"})
        self.n = n
        self.x1 = x1
        self.x2 = x2
        self.y = y
        self.d = d
        self.data = data

        self.test = spotify_confidence.ZTestLinreg(
            data_frame=data,
            numerator_column="y_sum",
            numerator_sum_squares_column="y2_sum",
            denominator_column="y_count",
            categorical_group_columns=["variation_name"],
            feature_column="Xs",
            feature_sum_squares_column="XX",
            feature_cross_sum_column="Xy",
            interval_size=0.99,
            correction_method="bonferroni",
            metric_column="metric_name",
        )

    def test_summary(self):
        summary_df = self.test.summary(verbose=True)
        print(summary_df)
        assert len(summary_df) == len(self.data)

    def test_parameters_univariate(self):
        def linreg(X, y):
            return np.matmul(np.linalg.inv(np.matmul(np.transpose(X), X)), np.matmul(np.transpose(X), y))

        summary_df = self.test.summary(verbose=True)
        X = np.ones((self.n, 3))
        X[:, 1] = self.x1
        X[:, 2] = self.x2
        b0 = linreg(X, self.y)
        y_adj = self.y - np.matmul(X, b0)
        X = np.ones((self.n, 2))
        X[:, 1] = self.d
        b1 = linreg(X, y_adj)
        assert np.allclose(b0[1:], summary_df[REGRESSION_PARAM][0].reshape(1, 2), rtol=0.0001)

        diff = self.test.difference(level_1="0", level_2="1", groupby="metric_name", verbose=True)
        assert np.allclose(b1[1], diff["difference"])

        v0 = np.var(y_adj[self.d == 0])
        v1 = np.var(y_adj[self.d == 1])
        n0 = y_adj[self.d == 0].size
        n1 = y_adj[self.d == 1].size
        assert np.allclose(diff["std_err"], np.sqrt(v0 / n0 + v1 / n1), rtol=1e-3)


class TestMultivariateMultipleMetrics(object):
    def setup(self):
        np.random.seed(123)

        n = 10000
        d = np.random.randint(2, size=n)
        x1 = np.random.standard_normal(size=n)
        x2 = np.random.standard_normal(size=n)
        y = 0.5 * d + 0.5 * x1 + 0.5 * x2 + np.random.standard_normal(size=n)
        metric_name = np.random.choice(["metricA", "metricB"], n)
        df = pd.DataFrame(
            {"variation_name": list(map(str, d)), "y": y, "x1": x1, "x2": x2, "metric_name": metric_name}
        )

        data = (
            df.assign(y2=y**2)
            .groupby(["variation_name", "metric_name"])
            .agg({"y": ["sum", "count"], "y2": "sum"})
            .reset_index()
        )

        data.columns = data.columns.map("_".join).str.strip("_")

        def _to_XX(data, features):
            X = data[features].to_numpy()
            XX = np.matmul(np.transpose(X), X)
            return XX

        def _to_Xy(data, features, metric):
            X = data[features].to_numpy()
            y = data[metric].to_numpy()
            Xy = np.matmul(np.transpose(X), y)
            return Xy

        def _to_Xs(data, features):
            X = data[features].to_numpy()
            Xs = np.sum(X, axis=0)
            return Xs

        XX = (
            df.groupby(["variation_name", "metric_name"])
            .apply(lambda x: pd.Series({"out": _to_XX(x, features=["x1", "x2"])}))
            .to_records()["out"]
        )
        Xy = (
            df.groupby(["variation_name", "metric_name"])
            .apply(lambda x: pd.Series({"out": _to_Xy(x, features=["x1", "x2"], metric="y")}))
            .to_records()["out"]
        )
        Xs = (
            df.groupby(["variation_name", "metric_name"])
            .apply(lambda x: pd.Series({"out": _to_Xs(x, features=["x1", "x2"])}))
            .to_records()["out"]
        )

        data["XX"] = XX
        data["Xy"] = Xy
        data["Xs"] = Xs

        self.n = n
        self.x1 = x1
        self.x2 = x2
        self.y = y
        self.d = d
        self.data = data
        self.m = metric_name

        self.test = spotify_confidence.ZTestLinreg(
            data_frame=data,
            numerator_column="y_sum",
            numerator_sum_squares_column="y2_sum",
            denominator_column="y_count",
            categorical_group_columns=["variation_name"],
            feature_column="Xs",
            feature_sum_squares_column="XX",
            feature_cross_sum_column="Xy",
            interval_size=0.99,
            correction_method="bonferroni",
            metric_column="metric_name",
        )

    def test_summary(self):
        summary_df = self.test.summary(verbose=True)
        print(summary_df)
        assert len(summary_df) == len(self.data)

    def test_parameters_multivariate(self):
        def linreg(X, y):
            return np.matmul(np.linalg.inv(np.matmul(np.transpose(X), X)), np.matmul(np.transpose(X), y))

        summary_df = self.test.summary(verbose=True)
        N0 = self.m[self.m == "metricA"].size
        # N1 = self.n - N0
        X = np.ones((N0, 3))
        X[:, 1] = self.x1[self.m == "metricA"]
        X[:, 2] = self.x2[self.m == "metricA"]
        b0 = linreg(X, self.y[self.m == "metricA"])
        y_adj = self.y[self.m == "metricA"] - np.matmul(X, b0)
        X = np.ones((N0, 2))
        X[:, 1] = self.d[self.m == "metricA"]
        b1 = linreg(X, y_adj)
        assert np.allclose(b0[1], summary_df[REGRESSION_PARAM][0][0], rtol=0.0001)
        assert np.allclose(b0[2], summary_df[REGRESSION_PARAM][0][1], rtol=0.0001)

        diff = self.test.difference(level_1=("0", "metricA"), level_2=("1", "metricA"), verbose=True)
        assert np.allclose(b1[1], diff["difference"])

        summary_df = self.test.summary(verbose=True)
        N0 = self.m[self.m == "metricB"].size
        # N1 = self.n - N0
        X = np.ones((N0, 3))
        X[:, 1] = self.x1[self.m == "metricB"]
        X[:, 2] = self.x2[self.m == "metricB"]
        b0 = linreg(X, self.y[self.m == "metricB"])
        y_adj = self.y[self.m == "metricB"] - np.matmul(X, b0)
        X = np.ones((N0, 2))
        X[:, 1] = self.d[self.m == "metricB"]
        b1 = linreg(X, y_adj)
        assert np.allclose(b0[1], summary_df[REGRESSION_PARAM][1][0], rtol=0.0001)
        assert np.allclose(b0[2], summary_df[REGRESSION_PARAM][1][1], rtol=0.0001)

        diff = self.test.difference(level_1=("0", "metricB"), level_2=("1", "metricB"), verbose=True)
        assert np.allclose(b1[1], diff["difference"])


class TestUnivariateMultiMetricRequiredSampleSize(object):
    def setup(self):
        np.random.seed(123)
        n = 2000000
        d = np.random.randint(2, size=n)
        x = np.random.standard_normal(size=n)
        y = 0.5 * d + 0.5 * x + np.random.standard_normal(size=n)
        m = np.repeat(["metricA", "metricB"], int(n / 2))
        data = pd.DataFrame({"variation_name": list(map(str, d)), "metric_name": m, "y": y, "x": x})
        data = (
            data.assign(xy=y * x)
            .assign(x2=x**2)
            .assign(y2=y**2)
            .groupby(["variation_name", "metric_name"])
            .agg({"y": ["sum", "count"], "y2": "sum", "x": "sum", "x2": "sum", "xy": "sum"})
            .reset_index()
        )

        data.columns = data.columns.map("_".join).str.strip("_")
        data = data.assign(non_inferiority_margin=5.0).assign(preferred_direction=DECREASE_PREFFERED)

        self.n = n
        self.x = x
        self.y = y
        self.d = d
        self.m = m
        self.data = data

        self.test = spotify_confidence.ZTest(
            data_frame=data,
            numerator_column="y_sum",
            numerator_sum_squares_column="y2_sum",
            denominator_column="y_count",
            categorical_group_columns=["variation_name"],
            interval_size=0.99,
            correction_method="bonferroni",
            metric_column="metric_name",
        )

        self.test_linreg = spotify_confidence.ZTestLinreg(
            data_frame=data,
            numerator_column="y_sum",
            numerator_sum_squares_column="y2_sum",
            denominator_column="y_count",
            categorical_group_columns=["variation_name"],
            feature_column="x_sum",
            feature_sum_squares_column="x2_sum",
            feature_cross_sum_column="xy_sum",
            interval_size=0.99,
            correction_method="bonferroni",
            metric_column="metric_name",
        )

    def test_summary(self):
        summary_df = self.test.summary(verbose=True)
        print(summary_df)
        assert len(summary_df) == len(self.data)

    def test_parameters_univariate_required_sample_size(self):
        diff = self.test.difference(
            level_1=("0", "metricB"), level_2=("1", "metricB"), verbose=True, non_inferiority_margins=True
        )
        diff_linreg = self.test_linreg.difference(
            level_1=("0", "metricB"), level_2=("1", "metricB"), verbose=True, non_inferiority_margins=True
        )
        assert (
            np.ceil(diff["required_sample_size"][0] * diff_linreg["variance_1"][0] / diff["variance_1"][0])
            == diff_linreg["required_sample_size"][0]
        )


# TODO: Test for sequential data (w/ ordinal column)
# TODO: Test for segmentation
